theory DualModalOps
  imports WP_SP
begin

definition "dwp A P = {s . \<exists> s'. (s, s') \<in> A \<and> s' \<in> P}"

lemma dwp_wp_duality:
  "dwp A P = -\<langle>A\<rangle> (- P)"
  by(auto simp add: dwp_def wp_ext)

lemma dwp_emp:
  "dwp A {} = {}"
  by(auto simp add: dwp_def)

lemma dwp_emp_univ:
  "dwp A UNIV = UNIV \<longleftrightarrow> Domain A = UNIV"
  by(auto simp add: dwp_def)

lemma dwp_Id2:
  "dwp Id P = P"
  by(auto simp add: dwp_def)

lemma dwp_pred_mono:
  assumes "P \<subseteq> Q"
  shows "dwp A P \<subseteq> dwp A Q"
  using assms by(auto simp add: dwp_def wp_ext)

lemma dwp_act_mono:
  assumes "A \<subseteq> B"
  shows "dwp A P \<subseteq> dwp B P"
  using assms by(auto simp add: dwp_def wp_ext)

lemma dwp_int_pred_dist:
  shows "dwp A (P \<inter> Q) \<subseteq> dwp A P \<inter> dwp A Q"
  by(auto simp add: dwp_def wp_ext)

lemma dwp_Int_pred_dist:
  shows "dwp B (\<Inter> x \<in> A. P x) \<subseteq> (\<Inter> x \<in> A. dwp B (P x))"
  by(auto simp add: dwp_def wp_ext)

lemma dwp_un_pred_dist:
  shows "dwp B (P \<union> Q) = dwp B P \<union> dwp B Q"
  by(auto simp add: dwp_def wp_ext)

lemma dwp_Un_pred_dist:
  shows "dwp B (\<Union> x \<in> A. P x) = (\<Union> x \<in> A. dwp B (P x))"
  by(auto simp add: dwp_def wp_ext)

lemma Domain_dwp:
  assumes "Domain A \<subseteq> P"
  shows "dwp A P \<subseteq> P"
  using assms by(auto simp add: dwp_def wp_ext)

lemma dwp_diag_eq:
  "dwp (A O Q\<^sup>d) P = dwp A (P \<inter> Q)"
  by(auto simp add: dwp_def diag_def)

lemma dwp_diag_eq2:
  "Q \<inter> dwp (A) P = dwp (Q\<^sup>d O A) P"
  by(auto simp add: dwp_def diag_def)

lemma dwp_seq_comp:
  assumes "P \<subseteq> \<langle>B\<rangle> P"
  shows "dwp A (P \<inter> dwp B Q) \<subseteq> dwp (A O B) (P \<inter> Q)"
  using assms by(auto simp add: dwp_def wp_ext)

lemma dwp_seq_comp2:
  shows "dwp A (dwp B P) = dwp (A O B) P"
  by(auto simp add: dwp_def wp_ext)

lemma dwp_actual_possible:
  "Id \<subseteq> A \<Longrightarrow> P \<subseteq> dwp A P"
  by(auto simp add: dwp_def)

lemma dwp_wp_conj:
  assumes "P \<inter> Q \<subseteq> \<langle>B\<rangle> P'"
      and "P \<inter> Q \<subseteq> dwp B Q'"
    shows "P \<inter> Q \<subseteq> dwp B (P' \<inter> Q')"
  using assms
  by(fastforce simp add: dwp_def wp_ext)

lemma dwp_wp_push:
  assumes "\<And> y y'. (x, y) \<in> B \<and> (x, y') \<in> G \<Longrightarrow>
                                         (\<exists> z. (y, z) \<in> G \<and> (y', z) \<in> B)"
    and "x \<in> dwp B (\<langle>G\<rangle> P)"
  shows "x \<in> \<langle>G\<rangle> (dwp B P)"
  using assms apply(simp add: dwp_def wp_ext) by blast

lemma dwp_wp_push_pred:
  assumes "\<And> x. x \<in> X \<Longrightarrow> (\<forall> y y'. (x, y) \<in> G \<and> (x, y') \<in> B \<longrightarrow>
                                         (\<exists> z. (y, z) \<in> B \<and> (y', z) \<in> G))"
    and "X \<subseteq> \<langle>G\<rangle> X"
  shows "X \<inter> dwp B (\<langle>G\<rangle> P) \<subseteq> \<langle>G\<rangle> (X \<inter> dwp B P)"
  apply(rule wp_pred_int_pres)
  using assms apply blast
  apply rule
  apply(subgoal_tac "(\<And> y y'. (x, y) \<in> G \<and> (x, y') \<in> B \<longrightarrow>
                                         (\<exists> z. (y, z) \<in> B \<and> (y', z) \<in> G))")
  apply(simp add: dwp_def wp_ext) apply blast
  apply(subgoal_tac "x \<in> X") defer apply blast
  using assms by blast


lemma dwp_wp_push_pred2:
  assumes "\<And> x. x \<in> X \<Longrightarrow> (\<forall> y y'. (x, y) \<in> G \<and> (x, y') \<in> B \<longrightarrow>
                                         (\<exists> z. (y, z) \<in> B \<and> (y', z) \<in> G))"
    and "X \<subseteq> \<langle>G\<rangle> X"
    and "X \<inter> dwp B P \<subseteq> Q"
  shows "X \<inter> dwp B (\<langle>G\<rangle> P) \<subseteq> \<langle>G\<rangle> Q"
  using dwp_wp_push_pred
  using assms(1) assms(2) assms(3) subset_Un_eq wp_ir_pr_disj_intro1
  by metis

lemma dwp_wp_transfer:
  assumes "P \<subseteq> \<langle>B\<rangle> P"
      and "\<And> y'. (x, y') \<in> G' \<Longrightarrow> \<exists> y z. (x, y) \<in> G \<and> (y, z) \<in> B \<and> (y', z) \<in> B"
      and "x \<in> \<langle>G\<rangle> P"
    shows "x \<in> \<langle>G'\<rangle> (dwp B P)"
  using assms apply(simp add: dwp_def wp_ext) by blast


lemma
  assumes "\<And> x. x \<in> X \<Longrightarrow> (\<forall> y y'. (x, y) \<in> G \<and> (x, y') \<in> B \<longrightarrow>
                                         (\<exists> z. (y, z) \<in> B \<and> (y', z) \<in> G))"
    and "B O B = B"
    and "X \<subseteq> \<langle>G\<rangle> X"
   shows "X \<inter> dwp B (\<langle>G\<rangle> (dwp B P)) \<subseteq> \<langle>G\<rangle> (X \<inter> (dwp B P))"
  by (smt assms(1) assms(2) assms(3) dwp_seq_comp2 dwp_wp_push_pred)


lemma dwp_wp_transfer2:
  assumes "\<And> x. x \<in> X \<Longrightarrow> (\<forall> y y'. (x, y) \<in> G' \<and> (x, y') \<in> B \<longrightarrow>
                                         (\<exists> z. (y, z) \<in> B \<and> (y', z) \<in> G'))"
      and "B O B = B"
      and "X \<subseteq> \<langle>G'\<rangle> X"
      and  "P \<subseteq> \<langle>B\<rangle> P"
      and "X \<subseteq> \<langle>B\<rangle> X"
      and "\<And> x y'. x \<in> X \<Longrightarrow> (x, y') \<in> G' \<Longrightarrow> \<exists> y z. (x, y) \<in> G \<and> (y, z) \<in> B \<and> (y', z) \<in> B"
    shows "X \<inter> dwp B (\<langle>G\<rangle> P) \<subseteq> \<langle>G'\<rangle> (X \<inter> dwp B P)"
  apply(rule subset_trans[where B="X \<inter> dwp B (\<langle>G'\<rangle> (dwp B P))"])
  using assms(4, 5, 6) apply(auto simp add: dwp_def wp_ext)[1] apply blast
  by (smt assms(1) assms(2) assms(3) dwp_seq_comp2 dwp_wp_push_pred)


lemma dwp_wp_transfer_full:
  assumes "X \<subseteq> \<langle>G'\<rangle> X"
      and "X \<inter> dwp B P \<subseteq> Q"
      and "X \<inter> \<langle>G\<rangle> P \<subseteq> \<langle>G'\<rangle> (dwp B P)"
    shows "X \<inter> \<langle>G\<rangle> P \<subseteq> \<langle>G'\<rangle> Q"
  apply(rule subset_trans[where B= "\<langle>G'\<rangle> (X \<inter> dwp B P)"])
   apply (metis Int_lower1 assms(1) assms(3) dual_order.trans wp_ir_pr_conj_intro)
  by (simp add: assms(2) wp_pred_mono)


end