theory LoadBuffering
  imports COComp
begin

locale LoadBuffering = RWActions
begin

datatype PC = L1 | L2 | End

definition "T1 = (1 :: nat)"
definition "T2 = (2 :: nat)"

lemma T_neq [simp]: "T1 = T2 \<longleftrightarrow> False"
  by(auto simp add: T1_def T2_def)

definition "trd_var x y t \<equiv> if (t = T1) then x else y"
definition "twr_var x y t \<equiv> if (t = T1) then y else x"


fun load_buffering' :: " R \<Rightarrow> L \<Rightarrow> L \<Rightarrow> nat \<Rightarrow> PC \<Rightarrow> PC AComm"
  where
  "load_buffering' r x y t L1 = loadc False r (trd_var x y t) L2" |
  "load_buffering' r x y t L2 = storec False (twr_var x y t) 1 End"  |
  "load_buffering' r x y t End = skip"


definition outline2 :: "R \<Rightarrow> L \<Rightarrow> L \<Rightarrow> T \<Rightarrow> PC \<Rightarrow> (('b, PC) full_state) set"
  where
"outline2 r x y t pc \<equiv> 
              case pc of
                   L1 \<Rightarrow>  (if t = T1 then lift\<^sub>1(synced T2 y 0) \<inter> Reg T2 (reg r 0) else (if t = T2 then lift\<^sub>1(synced T1 x 0) \<inter> Reg T1 (reg r 0) else {})) |
                   L2 \<Rightarrow>  (if t = T1 then lift\<^sub>1(synced T2 y 0) \<inter> Reg T2 (reg r 0) else (if t = T2 then lift\<^sub>1(synced T1 x 0) \<inter> Reg T1 (reg r 0) else {})) |
                   End \<Rightarrow> (if t = T1 then Reg T1 (reg r 0) \<union> Reg T2 (reg r 0) else (if t = T2 then Reg T1 (reg r 0) \<union> Reg T2 (reg r 0) else {}))"

lemma outline2_arb_pc:
  "outline2 r x y t pc \<subseteq> \<langle>arb_pc t'\<rangle> (outline2 r x y t pc)"
  apply(cases pc)
    apply(auto simp add: Reg_def lift\<^sub>1_def lift\<^sub>2_def lift_t_def arb_pc_def wp_ext outline2_def synced_def)
   apply (metis (mono_tags, lifting) fst_conv less_numeral_extra(3))
  by (metis (mono_tags, lifting) fst_conv less_numeral_extra(3))



lemma local_correctness_load_buffering:
  "local_correctness {T1, T2} (load_buffering' r x y) (outline2 r x y) t"
  apply(rule local_correctness_cases)
  using outline2_arb_pc apply fastforce
     apply(case_tac pc; simp)
  apply(case_tac pc; simp)
   apply(case_tac pc; simp)
  apply(simp add: outline2_def) 
  apply (smt Int_lower2 RWActions.load_wp_\<beta>_stable RWActions_axioms RWVars.synced_value_\<beta>_stable RWVarsActions.act_wp_Reg_stable_other RWVarsActions_axioms RWVars_axioms T_neq le_infI1 subset_trans wp_ir_pr_conj_intro)
   apply(case_tac pc; simp)
  apply(simp add: outline2_def)
  apply (smt RWVarsActions.act_wp_Reg_stable_other RWVarsActions_axioms T_neq inf_sup_absorb le_inf_iff wp_ir_pr_conj_elim2 wp_ir_pr_disj_intro2 wp_pred_int_distrib)
   apply(case_tac pc; simp)
   apply(simp add: outline2_def) 
   apply(case_tac pc; simp)   apply(simp add: outline2_def) 
   by(case_tac pc; simp)

lemma noninterference_load_buffering:
  assumes "x \<noteq> y"
  shows  "noninterference {T1, T2} (load_buffering' r x y) (outline2 r x y) t"
  apply(rule noninterference_cases)
  using outline2_arb_pc apply fastforce
     apply(case_tac pc; simp)
     apply(simp add: outline2_def)
  apply(case_tac pc'; simp)
     apply (intro conjI impI; elim conjE; simp add: trd_var_def)
         apply(rule wp_ir_pr_conj_intro) 
       apply (meson Int_lower1 load_wp_\<beta>_stable subset_trans synced_value_\<beta>_stable)
 apply(case_tac "v = 0")
  apply (simp add: le_infI1 load_wp_Reg load_wp_\<beta>_stable synced_value_\<beta>_stable wp_pred_int_distrib)
  apply(rule load_wp_Reg_neg)
  apply(rule subset_trans[where B="lift\<^sub>1 (synced T2 y 0)"]) apply blast
  apply(rule lift\<^sub>1_mono) using R_Rds
  apply (smt RWVars.initial_not_value RWVars_axioms basic_trans_rules(31) mem_Collect_eq subsetI wp_emp_not_trans wp_ext)  
        apply(rule wp_ir_pr_conj_intro) 
          apply (meson Int_lower1 load_wp_\<beta>_stable subset_trans synced_value_\<beta>_stable)
 apply(case_tac "v = 0")
  apply (simp add: le_infI1 load_wp_Reg load_wp_\<beta>_stable synced_value_\<beta>_stable wp_pred_int_distrib)
  apply(rule load_wp_Reg_neg)
  apply(rule subset_trans[where B="lift\<^sub>1 (synced T1 x 0)"]) apply blast
  apply(rule lift\<^sub>1_mono) using R_Rds
  apply (smt RWVars.initial_not_value RWVars_axioms basic_trans_rules(31) mem_Collect_eq subsetI wp_emp_not_trans wp_ext)    
     apply(simp add: outline2_def)
  apply(intro impI conjI; case_tac "pc'"; simp add: trd_var_def; elim conjE)
     apply(rule wp_ir_pr_conj_intro) 
          apply (meson Int_lower1 load_wp_\<beta>_stable subset_trans synced_value_\<beta>_stable)
 apply(case_tac "v = 0")
  apply (simp add: le_infI1 load_wp_Reg load_wp_\<beta>_stable synced_value_\<beta>_stable wp_pred_int_distrib)
  apply(rule load_wp_Reg_neg)
  apply(rule subset_trans[where B="lift\<^sub>1 (synced T2 y 0)"]) apply blast
  apply(rule lift\<^sub>1_mono) using R_Rds
  apply (smt RWVars.initial_not_value RWVars_axioms basic_trans_rules(31) mem_Collect_eq subsetI wp_emp_not_trans wp_ext)    
     apply(rule wp_ir_pr_conj_intro) 
          apply (meson Int_lower1 load_wp_\<beta>_stable subset_trans synced_value_\<beta>_stable)
 apply(case_tac "v = 0")
  apply (simp add: le_infI1 load_wp_Reg load_wp_\<beta>_stable synced_value_\<beta>_stable wp_pred_int_distrib)
  apply(rule load_wp_Reg_neg)
  apply(rule subset_trans[where B="lift\<^sub>1 (synced T1 x 0)"]) apply blast
  apply(rule lift\<^sub>1_mono) using R_Rds
  apply (smt RWVars.initial_not_value RWVars_axioms basic_trans_rules(31) mem_Collect_eq subsetI wp_emp_not_trans wp_ext)      
   apply(simp add: outline2_def)
   apply(intro impI conjI; case_tac "pc'"; simp add: trd_var_def; elim conjE)
   apply (simp add: act_wp_Reg_stable_other inf.coboundedI2 wp_ir_pr_disj_intro1)
    apply (meson Int_lower2 T_neq act_wp_Reg_stable_other subset_trans wp_ir_pr_disj_intro2)
   apply(simp add: outline2_def)
  apply(intro impI conjI; case_tac "pc'"; simp add: trd_var_def; elim conjE; case_tac "pc"; simp add: twr_var_def) 
     apply(rule wp_ir_pr_conj_intro) 
  apply (metis assms RWVars.synced_value_\<beta>_stable RWVars_axioms W_var  interf_wp2 le_infI1 object_write_lifted_wp synced_value_stable)
    using store_wp_Reg_stable apply fastforce
       apply(rule wp_ir_pr_conj_intro) 
  apply (metis assms RWVars.synced_value_\<beta>_stable RWVars_axioms W_var  interf_wp2 le_infI1 object_write_lifted_wp synced_value_stable)
    using store_wp_Reg_stable apply fastforce
    apply (simp add: act_wp_Reg_stable_other inf.coboundedI2 wp_ir_pr_disj_intro1)
       apply(rule wp_ir_pr_conj_intro) 
  apply (metis assms RWVars.synced_value_\<beta>_stable RWVars_axioms W_var  interf_wp2 le_infI1 object_write_lifted_wp synced_value_stable)
    using store_wp_Reg_stable apply fastforce
       apply(rule wp_ir_pr_conj_intro) 
  apply (metis assms RWVars.synced_value_\<beta>_stable RWVars_axioms W_var  interf_wp2 le_infI1 object_write_lifted_wp synced_value_stable)
    using store_wp_Reg_stable apply fastforce
      apply (simp add: act_wp_Reg_stable_other wp_ir_pr_conj_elim2 wp_ir_pr_disj_intro2)
    apply(case_tac pc'; simp) apply(case_tac pc'; simp)
    by(case_tac pc'; simp) 


definition "lb_init r x y \<equiv> lift\<^sub>1 (synced T1 x 0) \<inter>  lift\<^sub>1 (synced T2 x 0) \<inter> 
                                lift\<^sub>1 (synced T1 y 0) \<inter>  lift\<^sub>1 (synced T2 y 0)
                                     \<inter> Reg T1 (reg r 0) \<inter> Reg T2 (reg r 0) \<inter>
  {s . thr_pc s T1 = L1} \<inter> {s . thr_pc s T2 = L1}"

theorem
  assumes "x \<noteq> y"
  shows "reach (\<Union> e. full_step {T1, T2} (load_buffering' r x y) e) (lb_init r x y) \<subseteq> inv_of {T1, T2} (outline2 r x y)"
  apply(rule owicki_gries[OF local_correctness_load_buffering noninterference_load_buffering])
  using assms apply blast
  apply (simp add: lb_init_def outline2_def)
  apply rule
  apply(intro conjI impI)
  by simp


end
