theory MessagePassing
  imports COComp
begin

locale MessagePassing = MPSync
begin

datatype PC = L1 | L2 | L3 | End

definition "T1 = (1 :: nat)"
definition "T2 = (2 :: nat)"

lemma T_neq [simp]: "T1 = T2 \<longleftrightarrow> False"
  by(auto simp add: T1_def T2_def)

definition "trd_var x y t \<equiv> if (t = T1) then x else y"
definition "twr_var x y t \<equiv> if (t = T1) then y else x"


fun mp :: " R \<Rightarrow> R \<Rightarrow>  L \<Rightarrow> L \<Rightarrow> nat \<Rightarrow> PC \<Rightarrow> PC AComm"
  where
  "mp r1 r2 d f t L1 = (if t = T1 then storec False d 5 L2 else
                        if t = T2 then loadc True r1 f L2 else skip)" |
  "mp r1 r2 d f t L2 = (if t = T1 then storec True f 1 End else
                        if t = T2 then test (reg r1 1) L3 L1 else skip)"  |
  "mp r1 r2 d f t L3 = (if t = T1 then skip else
                        if t = T2 then loadc False r2 d End else skip)"  |
  "mp r1 r2 d f t End = skip"



definition outline :: "R \<Rightarrow> R \<Rightarrow> L \<Rightarrow> L \<Rightarrow> T \<Rightarrow> PC \<Rightarrow> (('b, PC) full_state) set"
  where
"outline r1 r2 d f t pc \<equiv> 
              case pc of
                   L1 \<Rightarrow>  (if t = T1 then lift\<^sub>1(synced T1 d 0) \<inter> lift\<^sub>1 (not_value T2 f 1)  else
                          (if t = T2 then lift\<^sub>1((\<langle>G (T2, R True f 1)\<rangle> (synced T2 d 5))) else {})) |
                   L2 \<Rightarrow>  (if t = T1 then lift\<^sub>1(synced T1 d 5) \<inter> lift\<^sub>1 (not_value T2 f 1) else
                          (if t = T2 then lift\<^sub>1((\<langle>G (T2, R True f 1)\<rangle> (synced T2 d 5))) \<inter> ((Reg t (-reg r1 1)) \<union> lift\<^sub>1(synced T2 d 5)) else {})) |
                   L3 \<Rightarrow>  (if t = T1 then UNIV else 
                          (if t = T2 then lift\<^sub>1(synced T2 d 5) else {})) |
                   End \<Rightarrow> (if t = T1 then UNIV else 
                          (if t = T2 then Reg T2 (reg r2 5) else {}))"

lemma outline_arb_pc:
  "outline  r1 r2 d f  t pc \<subseteq> \<langle>arb_pc t'\<rangle> (outline r1 r2 d f  t pc)"
  apply(cases pc)
    apply(auto simp add: Reg_def lift\<^sub>1_def lift\<^sub>2_def lift_t_def arb_pc_def wp_ext outline_def synced_def)
     apply (metis (mono_tags, lifting) fst_conv)
  by (metis (mono_tags, lifting) fst_conv)


lemma local_correctness:
  assumes "d \<noteq> f"
  shows "local_correctness  {T1, T2} (mp r1 r2 d f) (outline r1 r2 d f) t"
  apply(rule local_correctness_cases)
  using outline_arb_pc apply fastforce
     apply(simp add: outline_def) 
     apply(case_tac "t"; case_tac "pc"; simp)
  apply auto[1]
         apply(case_tac "pc_t"; simp)
  apply (simp add: T1_def T2_def)
  apply (simp add: T1_def T2_def inf_sup_aci(2) le_infI2 numeral_2_eq_2)
  apply rule 
  apply (simp add: T2_def)  
  apply (simp add: T2_def) 
  apply (simp add: T1_def T2_def)  
       apply auto[1]
   apply rule 
   apply rule 
   apply rule 
        apply auto[1]
       apply rule
       apply(case_tac "pc_t"; simp)
  apply (simp add: Int_assoc le_infI2 reg_neg_P)
       apply(case_tac "pc_t"; simp)
  apply auto[1]   
      apply auto[1]
  apply auto[1]
    apply(simp add: outline_def) 
  apply(case_tac t; simp)
  apply (metis (no_types, lifting) PC.case_distrib T1_def T2_def inf.orderI inf_bot_right inf_commute le_numeral_extra(2) le_numeral_extra(4) zero_neq_numeral)
    apply(intro conjI impI; case_tac pc; simp)
    apply blast
    apply(simp add: outline_def) 
  apply(case_tac t; simp)
  apply (metis (no_types, hide_lams) Int_empty_left PC.case_distrib T1_def T2_def empty_subsetI one_eq_numeral_iff zero_neq_numeral)
    apply(intro conjI impI; case_tac pc; simp)
         apply(rule wp_ir_pr_conj_intro) 
  apply (simp add: G_\<beta>_stable load_wp_\<beta>_stable synced_value_\<beta>_stable)
  apply(elim conjE)
  apply(case_tac "v = 1")
    apply(rule wp_ir_pr_disj_intro2)
  apply (simp add: object_read_lifted_wp)
    apply(rule wp_ir_pr_disj_intro1)
  apply (simp add: load_wp_Reg3)
  apply(case_tac "v = 5") 
    apply (simp add: load_wp_Reg)
  apply(rule load_wp_Reg_neg)
  apply(rule subset_trans[where B="lift\<^sub>1 (synced T2 d 5)"]) apply blast
  apply(rule lift\<^sub>1_mono) using R_Rds
  apply (smt RWVars.initial_not_value RWVars_axioms basic_trans_rules(31) mem_Collect_eq subsetI wp_emp_not_trans wp_ext)      
    apply(simp add: outline_def) 
  apply(case_tac t; simp)  
  apply (metis (no_types, hide_lams) Int_empty_left PC.case_distrib T1_def T2_def empty_subsetI one_eq_numeral_iff zero_neq_numeral)
    apply(intro conjI impI; case_tac pc; simp)  
  using T_neq apply auto[1] 
  apply(elim conjE)
         apply(rule wp_ir_pr_conj_intro) 
  using assms 
  apply (simp add: le_infI1 object_write_lifted_wp synced_value_self_write)
  using not_value_stable
  apply (simp add: inf.coboundedI1 inf_commute no_value_stable_diff_var_store)
    apply(case_tac pc; cases "t = T1"; cases "t = T2"; clarsimp)
apply(case_tac pc; cases "t = T1"; cases "t = T2"; clarsimp)
  by(case_tac pc; cases "t = T1"; cases "t = T2"; clarsimp)



lemma noninterference:
  assumes "d \<noteq> f"
  shows  "noninterference  {T1, T2} (mp r1 r2 d f) (outline r1 r2 d f) t"
  apply(rule noninterference_cases)
  using outline_arb_pc apply fastforce
     apply(simp add: outline_def)
   apply(case_tac "t")
    apply(intro conjI impI; case_tac pc; case_tac pc'; simp)
         apply(rule wp_ir_pr_conj_intro) 
  apply (meson Int_lower1 R_wval object_read_lifted_wp read_preserves_all subset_trans synced_value_\<beta>_stable)
  using assms
  apply(elim conjE, simp)
  using  load_wp_\<beta>_stable synced_value_\<beta>_stable
  
  apply (smt Int_iff in_mono not_value_\<beta>_stable subsetI)
      apply(elim conjE) 
        apply(rule wp_ir_pr_conj_intro) 
  using  load_wp_\<beta>_stable synced_value_\<beta>_stable
       apply (smt Int_lower1 subset_trans)
  using  not_value_\<beta>_stable subsetI
  apply (meson Int_iff load_wp_\<beta>_stable subset_iff)
        apply(rule wp_ir_pr_conj_intro) 
  using  load_wp_\<beta>_stable synced_value_\<beta>_stable
      apply (smt Int_lower1 subset_trans)
  using  not_value_\<beta>_stable subsetI
     apply (meson Int_iff load_wp_\<beta>_stable subset_iff)
        apply(rule wp_ir_pr_conj_intro) 
  using  load_wp_\<beta>_stable synced_value_\<beta>_stable
      apply (smt Int_lower1 subset_trans)
  using  not_value_\<beta>_stable subsetI
     apply (meson Int_iff load_wp_\<beta>_stable subset_iff)
    apply(intro conjI impI; case_tac pc; case_tac pc'; simp)
        apply(rule wp_ir_pr_conj_intro) 

    using  load_wp_\<beta>_stable synced_value_\<beta>_stable
         apply (smt Int_lower1 subset_trans)
    apply (meson Int_iff load_wp_\<beta>_stable not_value_\<beta>_stable subset_iff)
          apply(rule wp_ir_pr_conj_intro) 

    using  load_wp_\<beta>_stable synced_value_\<beta>_stable
         apply (smt Int_lower1 subset_trans)
    apply (meson Int_iff load_wp_\<beta>_stable not_value_\<beta>_stable subset_iff)
            apply(rule wp_ir_pr_conj_intro) 

    using  load_wp_\<beta>_stable synced_value_\<beta>_stable
         apply (smt Int_lower1 subset_trans)
    apply (meson Int_iff load_wp_\<beta>_stable not_value_\<beta>_stable subset_iff)
            apply(rule wp_ir_pr_conj_intro) 

    using  load_wp_\<beta>_stable synced_value_\<beta>_stable
         apply (smt Int_lower1 subset_trans)
    apply (meson Int_iff load_wp_\<beta>_stable not_value_\<beta>_stable subset_iff)

         apply(simp add: outline_def)
   apply(case_tac "t")
     apply(intro conjI impI; case_tac pc; case_tac pc'; simp)
            apply (simp_all add: T2_def)
    apply(intro conjI impI; case_tac pc; case_tac pc'; simp)
            apply(rule wp_ir_pr_conj_intro) 
                  apply (simp_all add: T2_def)
    defer   
    apply (smt Int_assoc Int_lower1 assms lifted_transfer_intro subset_trans)
            apply(rule wp_ir_pr_conj_intro) 
           defer
           defer
            apply(rule wp_ir_pr_conj_intro) 
    using assms lifted_transfer_intro[where t=T1 and y=d]
            apply blast
           defer
    apply(subgoal_tac "synced T1 x 0 \<inter> synced T1 x 5 = {}")
  apply (smt Int_absorb Int_assoc Int_commute Int_empty_right empty_subsetI lift\<^sub>1_inter_empty_mono subset_empty synced_diff_threads_vals)
 
             apply (metis Int_assoc Int_commute Int_empty_right def_value_unique synced_def zero_neq_numeral)
    apply (metis RWVars.synced_value_\<beta>_stable RWVars_axioms W_var assms interf_wp le_infI1 object_write_lifted_wp synced_value_stable)
    apply (meson Int_lower1 store_wp_Reg_stable subset_trans)      
           apply (meson Int_lower1 act_wp_Reg_stable_other subset_trans)
    apply (simp add: outline_def)
    apply(case_tac pc; case_tac pc'; cases "t = T1"; cases "t = T2"; simp)
    apply auto[18]
       defer
       defer 
    apply (metis MPSync.transfer_wr_diff_var_pres MPSync_axioms W_var assms object_write_lifted_wp wp_ir_pr_conj_elim2)
   apply (smt MPSync.transfer_wr_diff_var_pres MPSync_axioms W_var assms object_write_lifted_wp wp_ir_pr_conj_elim2)
    apply(elim conjE)
    apply(subgoal_tac "synced 2 x 5 \<inter>
       synced T1 x 0 = {}")
    using lift\<^sub>1_inter_empty_mono
    apply auto[1]
    using synced_diff_val_diff_threads apply auto[1]
    apply (smt Int_iff RWVarsActions.store_wp_Reg_stable RWVarsActions_axioms inf_sup_absorb subsetD wp_pred_int_distrib)
    using synced_diff_val_diff_threads apply force
    apply (smt Un_subset_iff W_var assms interf_wp le_infI1 object_write_lifted_wp store_wp_Reg_stable synced_value_\<beta>_stable synced_value_stable wp_ir_pr_conj_elim5 wp_ir_pr_disj_intro1 wp_ir_pr_disj_intro2)
   apply(case_tac pc; case_tac pc'; case_tac "t'=T2"; case_tac "t'=T1"; simp)
   apply(case_tac pc; case_tac pc'; case_tac "t'=T2"; case_tac "t'=T1"; simp)
    done


definition "lb_init r1 r2 d f \<equiv> lift\<^sub>1 (synced T1 d 0) \<inter>  lift\<^sub>1 (synced T2 d 0) \<inter> 
                                lift\<^sub>1 (synced T1 f 0) \<inter>  lift\<^sub>1 (synced T2 f 0)
                                     \<inter> Reg T1 (reg r1 0) \<inter> Reg T2 (reg r1 0) 
                                     \<inter> Reg T1 (reg r2 0) \<inter> Reg T2 (reg r2 0) \<inter>
  {s . thr_pc s T1 = L1} \<inter>  {s . thr_pc s T2 = L1} \<inter> lift\<^sub>1((\<langle>G (T2, R True f 1)\<rangle> (synced T2 d 5)))"

theorem
  assumes "d \<noteq> f"
  shows "reach (\<Union> e. full_step {T1, T2} (mp r1 r2 d f) e) (lb_init r1 r2 d f) \<subseteq> inv_of {T1, T2} (outline r1 r2 d f)"
  apply(rule owicki_gries[OF local_correctness noninterference])
  using assms apply blast
  using assms apply blast
  apply (simp add: lb_init_def outline_def)
   apply rule
   apply rule
   apply rule
  using T_neq apply auto[1]
  apply(intro conjI impI)
  apply (metis Suc_1 T1_def T2_def in_mono initial_not_value lift\<^sub>1_mono not_value_def numeral_1_eq_Suc_0 numerals(1))
  by simp



end
