theory PairOps
  imports Main
begin


(* Liftings *)

definition "on\<^sub>1 r = {((x, y), (x', y)) | x x' y. (x, x') \<in> r}"
definition "on\<^sub>2 r = {((x, y), (x, y')) | x y y'. (y, y') \<in> r}"

lemma on\<^sub>1_emp [simp]: "on\<^sub>1 {} = {}"
  by(auto simp add: on\<^sub>1_def)

lemma on\<^sub>2_emp [simp]: "on\<^sub>2 {} = {}"
  by(auto simp add: on\<^sub>2_def)

lemma on\<^sub>1_int:
  "on\<^sub>1 r \<inter> on\<^sub>1 s = on\<^sub>1 (r \<inter> s)"
  by(auto simp add: on\<^sub>1_def)

lemma on\<^sub>1_un:
  "on\<^sub>1 r \<union> on\<^sub>1 s = on\<^sub>1 (r \<union> s)"
  by(auto simp add: on\<^sub>1_def)

lemma on\<^sub>2_int:
  "on\<^sub>2 r \<inter> on\<^sub>2 s = on\<^sub>2 (r \<inter> s)"
  by(auto simp add: on\<^sub>2_def)

lemma on\<^sub>2_un:
  "on\<^sub>2 r \<union> on\<^sub>2 s = on\<^sub>2 (r \<union> s)"
  by(auto simp add: on\<^sub>2_def)

lemma on\<^sub>2_Un:
  "(\<Union> a. on\<^sub>2 (r a)) = on\<^sub>2 (\<Union> a. r a)"
  by(auto simp add: on\<^sub>2_def)

lemma on\<^sub>2_Int:
  "(\<Inter> a. on\<^sub>2 (r a)) = on\<^sub>2 (\<Inter> a. r a)"
  by(auto simp add: on\<^sub>2_def)

lemma on\<^sub>1_conv:
  "(on\<^sub>1 r)\<inverse> = on\<^sub>1 (r\<inverse>)"
  by(auto simp add: on\<^sub>1_def)

lemma on\<^sub>2_conv:
  "(on\<^sub>2 r)\<inverse> = on\<^sub>2 (r\<inverse>)"
  by(auto simp add: on\<^sub>2_def)

lemma on\<^sub>1_Id [simp]:
  "on\<^sub>1 Id = Id"
  by(auto simp add: on\<^sub>1_def)

lemma on\<^sub>2_Id [simp]:
  "on\<^sub>2 Id = Id"
  by(auto simp add: on\<^sub>2_def)

lemma on\<^sub>1_mono: "r \<subseteq> s \<Longrightarrow> on\<^sub>1 r \<subseteq> on\<^sub>1 s"
  by(auto simp add: on\<^sub>1_def)

lemma on\<^sub>2_mono: "r \<subseteq> s \<Longrightarrow> on\<^sub>2 r \<subseteq> on\<^sub>2 s"
  by(auto simp add: on\<^sub>2_def)

lemma on\<^sub>1_distrib: "on\<^sub>1 r O on\<^sub>1 s = on\<^sub>1 (r O s)"
  by(auto simp add: on\<^sub>1_def)

lemma on\<^sub>2_distrib: "on\<^sub>2 r O on\<^sub>2 s = on\<^sub>2 (r O s)"
  by(auto simp add: on\<^sub>2_def)

lemma on\<^sub>1_on\<^sub>2_commute: "on\<^sub>1 r O on\<^sub>2 s = on\<^sub>2 s O on\<^sub>1 r"
  by(auto simp add: on\<^sub>2_def on\<^sub>1_def)

lemma on\<^sub>1_on\<^sub>2_ext:
  "on\<^sub>1 r O on\<^sub>2 s = {((x, y), (x', y')) | x y x' y'. (x, x') \<in> r \<and> (y, y') \<in> s}"
  by(auto simp add: on\<^sub>2_def on\<^sub>1_def)

definition "flift i r = {(f, f(i := x')) | f x'. (f i, x') \<in> r}"

lemma flift_emp [simp]: "flift i {} = {}"
  by(auto simp add: flift_def)

lemma flift_int:
  "flift t r \<inter> flift t s = flift t (r \<inter> s)"
  apply(simp add: flift_def)
  using fun_upd_eqD by fastforce

lemma flift_un:
  "flift t r \<union> flift t s = flift t (r \<union> s)"
  apply(simp add: flift_def)
  using fun_upd_eqD by fastforce

lemma flift_conv:
  "(flift t r)\<inverse> = flift t (r\<inverse>)"
  by(fastforce simp add: flift_def)

lemma flift_Id:
  "flift i Id = Id"
  by(auto simp add: flift_def)

lemma flift_mono:
  "r \<subseteq> s \<Longrightarrow> flift i r \<subseteq>  flift i s"
  by(fastforce simp add: flift_def)

lemma flift_commute:
  "i \<noteq> j \<Longrightarrow> flift i r O flift j s = flift j s O flift i r"
  by(fastforce simp add: flift_def)

lemma flift_distrib:
  "flift i r O flift i s = flift i (r O s)"
  by(fastforce simp add: flift_def)

definition "get v \<equiv> {(v, v)}"

lemma get_no_match [simp]: "u \<noteq> v \<Longrightarrow> get u O get v = {}"
  by(auto simp add: get_def)

lemma get_collapse [simp]: "get v O get v = get v"
  by(auto simp add: get_def)

definition "put v \<equiv> {(u, v) | u. True}"

lemma [simp]: "put u O put v = put v"
  by(auto simp add: put_def)

lemma [simp]: "put v O get v = put v"
  by(auto simp add: get_def put_def)

lemma [simp]: "u \<noteq> v \<Longrightarrow> put u O get v = {}"
  by(auto simp add: get_def put_def)

lemma continuation:
  "(x, y) \<in> (\<Union> v. get v O r v) \<longleftrightarrow> (x, y) \<in> r x"
  by(auto simp add: get_def)

lemma get_put_r:
  "(v, v') \<in> r \<Longrightarrow> get v O put v' \<subseteq> r"
  by(auto simp add: get_def put_def)

definition "lte = {(x :: ('a :: order), y) | x y. x \<le> y}"

lemma lte_test_lemma:
    "(f, g) \<in> lte \<Longrightarrow> f x \<le> g x"
  by (smt CollectD fst_conv le_funE lte_def snd_conv)

lemma lte_refl: "Id \<subseteq> lte"
  by(auto simp add: lte_def)

lemma lte_trans [simp]: "lte O lte = lte"
  by(auto simp add: lte_def)

lemma lte_antisym [simp]: "lte \<inter> (lte\<inverse>) = Id"
  by(auto simp add: lte_def)


definition liftR :: "'t \<Rightarrow> ('s \<times> 'v) rel \<Rightarrow> ('s \<times> ('t \<Rightarrow> 'v)) rel"
  where
  "liftR t R = {((s, vs), (s', vs(t := v'))) | s vs s' v'. ((s, vs(t)), (s', v')) \<in> R}"



lemma double_commute:
  assumes "r O t \<subseteq> t O r'"
      and "s O t \<subseteq> t O s'"
  shows "r O s O t \<subseteq> t O r' O s'"
  using assms by force

lemma Int_range_ext:
    "(\<Inter> b. r b) = {x . (\<forall> b. x \<in> r b)}"
  by blast



lemma skolem_relcompI:
  assumes "(x, z) \<in> R' a b O S"
    and "\<And> y. (x, y) \<in> R' a b \<Longrightarrow> (y, z) \<in> S \<Longrightarrow> (f a x y z, z) \<in> R"
    and "\<And> y. (x, y) \<in> R' a b \<Longrightarrow> (y, z) \<in> S \<Longrightarrow> (x, f a x y z) \<in> S"
  shows "(x, z) \<in> S O R"
  using assms(1) assms(2) assms(3) by blast

lemma pointwise_commute1:
  assumes "\<And> y. (x, y) \<in> (\<Inter> b. R b) \<Longrightarrow> (y, z) \<in> r \<Longrightarrow>
                      (x, f x y z) \<in> r \<and> (f x y z, z) \<in> (\<Inter> b. R b)"
  and "(x, z) \<in> (\<Inter> b. R b) O r"
  shows "(x, z) \<in> r O (\<Inter> b. R b)"
  apply(rule relcompE[OF assms(2)], simp)
  by (meson assms(1) assms(2) relcomp.simps)

lemma pointwise_commute2:
  assumes "\<And> x y z. (\<forall> b. (x, y) \<in> R b) \<Longrightarrow> (y, z) \<in> r \<Longrightarrow>
                      (x, f x y z) \<in> r \<and> (\<forall> b. (f x y z, z) \<in> R b)"
  shows "(\<Inter> b. R b) O r \<subseteq> r O (\<Inter> b. R b)"
  apply(rule subrelI)
  apply(rule pointwise_commute1[where f=f]) using assms Int_range_ext by simp_all



(*
lemma liftR_mono:
  assumes "R \<subseteq> S"
  shows "\<langle>t : R\<rangle> \<subseteq>\<langle>t : S\<rangle>"
  using assms by(fastforce simp add: liftR_def)

lemma liftR_emp [simp]:
  shows "\<langle>t : {}\<rangle> = {}"
  by(fastforce simp add: liftR_def)

lemma liftR_Id [simp]:
  shows "\<langle>t : Id\<rangle> = Id"
  by(fastforce simp add: liftR_def)

lemma liftR_un_distrib:
  shows "\<langle>t: R \<union> S\<rangle> = \<langle>t : R\<rangle> \<union> \<langle>t : S\<rangle>"
  by(fastforce simp add: liftR_def)

lemma liftR_on\<^sub>2:
  "\<langle>t : on\<^sub>2 r\<rangle> = on\<^sub>2(flift t r)"
  by(auto simp add: liftR_def on\<^sub>2_def flift_def)

lemma liftR_relcomp:
  "\<langle>t : R\<rangle> O \<langle>t : S\<rangle> = \<langle>t : R O S\<rangle>"
  by(fastforce simp add: liftR_def)

lemma liftR_commutes_other:
  "t \<noteq> t' \<Longrightarrow> \<langle>t : on\<^sub>2 R\<rangle> O \<langle>t' : S\<rangle> = \<langle>t' : S\<rangle> O \<langle>t : on\<^sub>2 R\<rangle>"
  by(fastforce simp add: liftR_def on\<^sub>2_def)

lemma on\<^sub>1_anonymous: "\<langle>t : on\<^sub>1 R\<rangle> = on\<^sub>1 R"
  by(auto simp add: liftR_def on\<^sub>1_def)


definition
  "lift_fst R = on\<^sub>1 {(s, s') | s v s' v'. ((s, v), (s', v')) \<in> R}"

lemma lift_fst_liftR:
  "lift_fst r O \<langle>t: s\<rangle> = \<langle>t: lift_fst r O s\<rangle>"
  apply(unfold lift_fst_def)
  using on\<^sub>1_anonymous liftR_relcomp by fastforce

lemma on\<^sub>1_liftR_relcomp_neq_ext:
 "on\<^sub>1 R O \<langle>t' : S\<rangle> = {((s, vs),(s'', vs(t' := v''))) | s vs s' s'' v''.
                  (s, s') \<in> R \<and> ((s', vs(t')), (s'', v'')) \<in> S}"
apply(simp add: liftR_def on\<^sub>1_def) by fastforce


lemma liftR_relcomp_neq_ext:
 "t \<noteq> t' \<Longrightarrow>
   \<langle>t : R\<rangle> O \<langle>t' : S\<rangle> = {((s, vs),(s'', vs(t := v', t' := v''))) | s vs s' v' s'' v''.
                  ((s, vs(t)), (s', v')) \<in> R \<and> ((s', vs(t')), (s'', v'')) \<in> S}"
apply(simp add: liftR_def) by fastforce

lemma liftR_commutes_self:
  assumes "R O S \<subseteq> S O R"
  shows "\<langle>t : R\<rangle> O \<langle>t : S\<rangle> \<subseteq> \<langle>t : S\<rangle> O \<langle>t : R\<rangle>"
  apply (unfold liftR_relcomp)
  by(rule liftR_mono[OF assms])

lemma on\<^sub>1_liftR_commutes:
  assumes "on\<^sub>1 R O S \<subseteq> S O on\<^sub>1 R"
  shows "on\<^sub>1 R O \<langle>t' : S\<rangle> \<subseteq> \<langle>t' : S\<rangle> O on\<^sub>1 R"
  apply(unfold on\<^sub>1_anonymous[symmetric, where t=t'] liftR_relcomp)
  apply(rule liftR_mono)
  by (metis assms)

lemma on\<^sub>1_liftR_commutes2:
  assumes "on\<^sub>1 R O S \<subseteq> S O on\<^sub>1 R"
  shows "\<langle>t : on\<^sub>1 R\<rangle> O \<langle>t' : S\<rangle> \<subseteq> \<langle>t' : S\<rangle> O \<langle>t : on\<^sub>1 R\<rangle>"
  apply(unfold on\<^sub>1_anonymous)
  using assms by (simp add: on\<^sub>1_liftR_commutes)


definition "copy t t' = {((s, v), (s, v(t' := v(t)))) | s v. True}"

lemma copy_Id [simp]:
  "copy t t = Id"
  by(fastforce simp add: copy_def)

lemma copy_liftR:
  "copy t t' O \<langle>t' : r\<rangle> =
        {((s, vs), (s', vs(t' := v'))) | s vs s' v'. ((s, vs(t)), (s', v')) \<in> r}"
  by(fastforce simp add: copy_def on\<^sub>2_def liftR_def)

lemma liftR_copy:
  "\<langle>t : r\<rangle> O copy t t' =
        {((s, vs), (s', vs(t := v', t' := v'))) | s vs s' v'. ((s, vs(t)), (s', v')) \<in> r}"
  by(fastforce simp add: copy_def on\<^sub>2_def liftR_def)

lemma copy_r_inclusion:
  "copy t t' O \<langle>t' : r\<rangle> O copy t' t = \<langle>t : r\<rangle> O copy t t'"
  by(fastforce simp add: liftR_copy liftR_def copy_def )

lemma copy_r_other1:
  "t1 \<noteq> t \<Longrightarrow> t2 \<noteq> t \<Longrightarrow> copy t1 t2 O \<langle>t : r\<rangle> \<subseteq> \<langle>t : r\<rangle> O copy t1 t2"
  apply(auto simp add: liftR_def copy_def)
  by fastforce

*)
end