theory Simulations
  imports PairOps WP_SP
begin


definition
  "sim G \<equiv> {r . \<forall> a. r O G a \<subseteq> G a O r}"

lemma sim_ext:
  "sim G = {r. \<forall> a x y z. (x, y) \<in> r \<and> (y, z) \<in> G a \<longrightarrow> (\<exists> y'. (x, y') \<in> G a \<and> (y', z) \<in> r)}"
  by(fastforce simp add: sim_def)

lemma sim_Id:
   "Id \<in> sim G"
  by(simp add: sim_def)

lemma sim_emp:
   "{} \<in> sim G"
  by(simp add: sim_def)

lemma sim_join_closed:
  assumes "\<And> r. r \<in> rs \<Longrightarrow> r \<in> sim G"
    shows "\<Union> rs \<in> sim G"
  using assms apply(simp add: sim_ext)
  by metis

definition "\<alpha> G = \<Union> (sim G)"

lemma \<alpha>_sim: "\<alpha> G \<in> sim G"
  apply(unfold \<alpha>_def)
  by(rule sim_join_closed, simp)

lemma \<alpha>_gt_sim:
  assumes "r \<in> sim G"
  shows "r \<subseteq> \<alpha> G"
  using assms by(fastforce simp add: \<alpha>_def)

lemma sim_commutes:
  assumes "r \<in> sim G"
  shows "r O G a \<subseteq> G a O r"
  using assms by(simp add: sim_def)

lemma \<alpha>_commutes:
  "\<alpha> G O G a \<subseteq> G a O \<alpha> G"
  by(rule sim_commutes[OF \<alpha>_sim])

lemma \<alpha>_refl:
  "Id \<subseteq> \<alpha> G"
  using \<alpha>_def sim_Id by fastforce

lemma \<alpha>_trans:
  "\<alpha> G O \<alpha> G = \<alpha> G"
  apply rule
  apply(subgoal_tac "\<alpha> G O \<alpha> G \<in> sim G")
    apply (simp add: Sup_upper \<alpha>_def)
  using \<alpha>_sim apply(simp add: sim_def O_assoc)
   apply rule apply(rule double_commute[OF \<alpha>_commutes \<alpha>_commutes])
  using \<alpha>_refl by fastforce


definition "bndsim G B = {A . \<forall> b \<in> B. b \<subseteq> \<langle>A\<rangle> b} \<inter> sim G"

lemma bndsim_join_closed:
  assumes "rs \<subseteq> bndsim G B"
    shows "\<Union> rs \<in> bndsim G B"
  using assms apply(simp add: bndsim_def)
  apply rule
  using wp_Sup_Int weakest_action_correct apply fastforce
  by (simp add: sim_join_closed subset_eq)

lemma bndsim_join_closed2:
  shows "\<Union> (bndsim G B) \<in> bndsim G B"
  using bndsim_join_closed
  by (simp add: bndsim_join_closed)

lemma bndsim_commutes:
  assumes "r \<in> bndsim G B"
  shows "r O G x \<subseteq> G x O r"
  using assms by(fastforce simp add: bndsim_def sim_def)


lemma bndsim_preserves:
  assumes "r \<in> bndsim G B"
      and "b \<in> B"
    shows "b \<subseteq> \<langle>r\<rangle> b"
  using assms by(auto simp add: bndsim_def)

lemma bndsim_intro:
  assumes "\<And> b. b \<in> B \<Longrightarrow> b \<subseteq> \<langle>r\<rangle> b"
    and "\<And> e. r O G e \<subseteq> G e O r"
  shows "r \<in> bndsim G B"
  using assms by(simp add: bndsim_def sim_def)

lemma bndsim_refl:
  "Id \<in> bndsim G B"
  apply(simp add: bndsim_def)
  by(rule sim_Id)

lemma bndsim_trans:
  assumes "r \<in> bndsim G B"
      and "s \<in> bndsim G B"
    shows "r O s \<in> bndsim G B"
  apply(rule bndsim_intro)
  using assms apply(simp add: bndsim_def) apply (metis wp_pres_compos2)
  using assms apply(simp add: bndsim_def)
  by (simp add: O_assoc double_commute sim_commutes)


definition "psim G X = {r . (\<forall> x \<in> X. r O G x \<subseteq> G x O r)}"

lemma psim_join_closed:
  assumes "rs \<subseteq> psim G X"
    shows "\<Union> rs \<in> psim G X"
  using assms by(fastforce simp add: psim_def)

lemma psim_commutes:
  assumes "r \<in> psim G X"
    and "x \<in> X"
  shows "r O G x \<subseteq> G x O r"
  using assms by(fastforce simp add: psim_def )

lemma psim_Id:
  shows "Id \<in> psim G X"
  by(fastforce simp add: psim_def)

lemma psim_trans:
  assumes "r \<in> psim G X"
    and "s \<in> psim G Y"
  shows "r O s \<in> psim G (X \<inter> Y)"
  using assms apply(simp add: psim_def)
  by (metis (no_types, lifting) IntD1 IntD2 O_assoc double_commute)

lemma sim_psim_eq:
  "sim G = psim G UNIV"
  by(simp add: sim_def psim_def)

lemma int_psim_sim_union:
  "psim G X \<inter> psim G Y = psim G (X \<union> Y)"
  apply(auto simp add: psim_def)
  by blast


definition "prop_sim G B X = {A . \<forall> b \<in> B. b \<subseteq> \<langle>A\<rangle> b} \<inter> psim G X"

lemma prop_sim_intro:
  assumes "\<And> b. b \<in> B \<Longrightarrow> b \<subseteq> \<langle>R\<rangle> b"
    and "\<And> x. x \<in> X \<Longrightarrow> R O G x \<subseteq> G x O R"
  shows "R \<in> prop_sim G B X"
  using assms by(simp add: prop_sim_def psim_def)

lemma prop_sim_elim1:
  assumes "R \<in> prop_sim G B X"
     and "b \<in> B"
  shows "b \<subseteq> \<langle>R\<rangle> b"
  using assms by(simp add: prop_sim_def psim_def)

lemma prop_sim_elim2:
  assumes "R \<in> prop_sim G B X"
     and "x \<in> X"
  shows "R O G x \<subseteq> G x O R"
  using assms by(simp add: prop_sim_def psim_def)

definition "wk_prop_sim G B X = \<Union> (prop_sim G B X)"

lemma wk_prop_sim_is_prop_sim:
   "wk_prop_sim G B X \<in> prop_sim G B X"
  apply (simp add: wk_prop_sim_def prop_sim_def)
  apply rule 
  apply (metis (mono_tags, lifting) Ball_Collect inf_le1 wp_Sup2)
  by (simp add: psim_join_closed)

lemma wk_prop_sim_weakest:
  assumes "\<And> b. b \<in> B \<Longrightarrow> b \<subseteq> \<langle>R\<rangle> b"
    and "\<And> x. x \<in> X \<Longrightarrow> R O G x \<subseteq> G x O R"
  shows "R \<subseteq> wk_prop_sim G B X"
  apply (simp add: wk_prop_sim_def)
  apply(rule Sup_upper)
  apply(rule prop_sim_intro) 
  apply (simp add: assms(1))
  by (simp add: assms(2))

lemma prop_sim_Id:
  shows "Id \<in> prop_sim G B X"
  by (simp add: prop_sim_intro)

lemma prop_sim_trans:
  assumes "r \<in> prop_sim G B X"
    and "s \<in> prop_sim G B Y"
  shows "r O s \<in> prop_sim G B (X \<inter> Y)"
  apply(rule prop_sim_intro)
  apply (meson assms(1) assms(2) prop_sim_elim1 wp_pres_compose)
  by (metis (no_types, lifting) IntD1 IntD2 O_assoc assms(1) assms(2) double_commute prop_sim_elim2)

lemma int_prop_sim_union:
  "prop_sim G B X \<inter> prop_sim G C Y = prop_sim G (B \<union> C) (X \<union> Y)"
  apply rule
  apply rule
  apply(rule prop_sim_intro)
  apply (metis IntD1 IntD2 Un_iff prop_sim_elim1)
   apply (metis IntD1 IntD2 Un_iff prop_sim_elim2)
  apply rule
  by (simp add: prop_sim_elim1 prop_sim_elim2 prop_sim_intro)

lemma prop_sim_refl:
  shows "Id \<subseteq> wk_prop_sim G B X"
  by (simp add: wk_prop_sim_weakest)

lemma wk_prop_sim_trans:
  shows "wk_prop_sim G B X O wk_prop_sim G B X \<subseteq> wk_prop_sim G B (X \<inter> Y)"
  using wk_prop_sim_weakest
  using prop_sim_trans[OF wk_prop_sim_is_prop_sim wk_prop_sim_is_prop_sim]
  by (smt IntD1 inf.idem prop_sim_elim1 prop_sim_elim2)

lemma wk_prop_sim_P_stable:
  assumes "b \<in> B"
  shows "b \<subseteq> \<langle>wk_prop_sim G B X\<rangle> b"
  by(rule prop_sim_elim1[OF wk_prop_sim_is_prop_sim assms(1)])

lemma wk_prop_sim_commutes:
  assumes "x \<in> X"
  shows "wk_prop_sim G B X O G x \<subseteq> G x O wk_prop_sim G B X"
  by(rule prop_sim_elim2[OF wk_prop_sim_is_prop_sim assms(1)])

lemma wk_int_prop_sim_union:
  "wk_prop_sim G B X \<inter> wk_prop_sim G C Y = wk_prop_sim G (B \<union> C) (X \<union> Y)"
  using int_prop_sim_union
  apply(simp add: wk_prop_sim_def) 
  apply rule defer
  apply blast oops

lemma wk_prop_sim_antimono1:
   "B \<subseteq> C \<Longrightarrow> wk_prop_sim G C X \<subseteq> wk_prop_sim G B X"
  apply(simp add: wk_prop_sim_def prop_sim_def psim_def wp_ext)
  apply clarsimp
  by (smt IntI mem_Collect_eq subset_eq)

lemma wk_prop_sim_antimono2:
  "X \<subseteq> Y \<Longrightarrow> wk_prop_sim G B Y \<subseteq> wk_prop_sim G B X"
  apply(simp add: wk_prop_sim_def prop_sim_def psim_def wp_ext)
  apply clarsimp
  by (smt IntI mem_Collect_eq subset_eq)

definition
  "protects G B X = {r . (\<forall> x \<in> X. r O G x \<subseteq> G x O (wk_prop_sim G B X)) \<and>
                           (\<forall> b \<in> B. b \<subseteq> \<langle>r\<rangle> b)}"

lemma protects_intro:
  assumes "\<And> x.  x\<in> X \<Longrightarrow> r O G x \<subseteq> G x O (wk_prop_sim G B X)"
    and "\<And> b. b \<in> B \<Longrightarrow> b \<subseteq> \<langle>r\<rangle> b"
  shows "r \<in> protects G B X"
  using assms by(auto simp add: protects_def)

lemma protects_elim1:
  assumes "r \<in> protects G B X"
   and "x \<in> X"
 shows "r O G x \<subseteq> G x O (wk_prop_sim G B X)"
  using assms by(auto simp add: protects_def)

lemma protects_elim2:
  assumes "r \<in> protects G B X"
   and "b \<in> B"
  shows "b \<subseteq> \<langle>r\<rangle> b"
  using assms by(auto simp add: protects_def)

lemma protects_down_closed:
  assumes "r \<in> protects G B X"
     and "s \<subseteq> r"
   shows "s \<in> protects G B X"
   apply(rule protects_intro)
  using assms apply(simp add: protects_def)
   apply blast
  using assms
  by (metis inf.absorb_iff2 protects_elim2 wp_ir_act_int1)

lemma protects_join_closed:
  assumes "rs \<subseteq> protects G B X"
  shows "\<Union> rs \<in> protects G B X"
   apply(rule protects_intro)
  using assms apply(simp add: protects_def) apply auto 
  apply blast
  by (meson assms protects_elim2 subsetD wp_Sup2)

lemma protects_flat: "r \<subseteq> \<Union> (protects G B X) \<longleftrightarrow> r \<in> protects G B X"
  apply rule
   apply(rule protects_intro)
    apply(auto simp add: protects_def)[1]
    apply blast
   apply (metis inf.absorb_iff2 protects_elim2 wp_Sup2 wp_ir_act_int1)
  by (simp add: Sup_upper)

lemma
  "C \<subseteq> B \<Longrightarrow> protects G B X \<subseteq> protects G C X"
  apply (simp add: protects_def)apply clarsimp apply rule apply clarsimp
   apply (smt relcomp.relcompI relcomp.simps subsetD wk_prop_sim_antimono1)
  by blast


lemma "wk_prop_sim G B X \<in> protects G B X"
  by (metis prop_sim_elim1 prop_sim_elim2 protects_intro wk_prop_sim_is_prop_sim)

definition "prot G B X =
                   {r . (\<forall> b \<in> B. b \<subseteq> \<langle>r\<rangle> b) \<and>
                        (\<forall> p x. p \<subseteq> \<langle>r\<rangle> p \<and> x \<in> X \<longrightarrow> \<langle>G x\<rangle> p \<subseteq> \<langle>r\<rangle> (\<langle>G x\<rangle> p))}"


lemma prot_intro:
  assumes "\<And> b. b \<in> B \<Longrightarrow> b \<subseteq> \<langle>r\<rangle> b"
    and  "\<And> p x. p \<subseteq> \<langle>r\<rangle> p \<and> x \<in> X \<longrightarrow> \<langle>G x\<rangle> p \<subseteq> \<langle>r\<rangle> (\<langle>G x\<rangle> p)"
  shows "r \<in> prot G B X"
  using assms by(simp add: prot_def)

lemma prot_elim1:
  assumes "r \<in> prot G B X"
    and  "b \<in> B"
  shows "b \<subseteq> \<langle>r\<rangle> b"
  using assms by(simp add: prot_def)


lemma prot_elim2:
  assumes "r \<in> prot G B X"
      and "p \<subseteq> \<langle>r\<rangle> p"
      and "x \<in> X"
    shows "\<langle>G x\<rangle> p \<subseteq> \<langle>r\<rangle> (\<langle>G x\<rangle> p)"
  using assms by(simp add: prot_def)

lemma prot_join_closed:
  assumes "rs \<subseteq> prot G B X"
  shows "\<Union> rs \<in> prot G B X"
  apply(rule prot_intro)
   apply (metis assms prot_elim1 subsetD wp_Sup2)
  apply rule
  apply(rule wp_Sup2)
  by (metis Sup_upper assms inf.absorb_iff1 inf.coboundedI2 prot_elim2 subsetD wp_act_antimono2)

lemma prot_antimono1:
  "B \<subseteq> C \<Longrightarrow> prot G C X \<subseteq> prot G B X"
  apply rule
  apply(rule prot_intro)
  using prot_elim1 apply blast
  by (simp add: prot_elim2)

lemma prot_antimono2:
  "X \<subseteq> Y \<Longrightarrow> prot G B Y \<subseteq> prot G B X"
  apply rule
  apply(rule prot_intro)
  using prot_elim1 apply blast
  by (simp add: prot_elim2 subsetD)
  
lemma prot_int_sup_distrib:
  shows "prot G B X \<inter> prot G C Y = prot G (B \<union> C) (X \<union> Y)"
  apply rule
  apply rule apply(rule prot_intro) 
  apply (meson IntD1 IntD2 Un_iff prot_elim1)
  apply (meson IntD1 IntD2 Un_iff prot_elim2)
  by (fastforce simp add: prot_elim1 prot_elim2 prot_intro)

lemma prot_compositional:
  assumes "r \<in> prot G B X"
    and "r \<in> prot G C Y"
  shows "r \<subseteq> \<Union> (prot G (B \<union> C) (X \<union> Y))"
  using prot_int_sup_distrib
  using assms(1) assms(2) by blast

lemma prot_compositional2:
  assumes "r \<subseteq> \<Union> (prot G B X)"
    and "r \<subseteq> \<Union> (prot G C Y)"
  shows "r \<subseteq> \<Union> (prot G (B \<union> C) (X \<union> Y))"
  using prot_int_sup_distrib
  using assms(1) assms(2) nitpick
oops

lemma prot_sim1:
  assumes "r \<in> bndsim G B"
  shows "r \<in> prot G B UNIV"
  apply(rule prot_intro)
  using assms bndsim_def apply fastforce
  using assms by (metis (no_types, lifting) bndsim_commutes wp_lipton)

lemma prot_sim2:
  assumes "r \<in> bndsim G B"
  shows "r \<subseteq> \<Union> (prot G B UNIV)"
  by (simp add: Sup_upper assms prot_sim1)

lemma prot_relcomp_compositional1:
  assumes "r \<in> prot G B X"
      and "\<And> b. b \<in> B \<Longrightarrow> b \<subseteq> \<langle>s\<rangle> b"
      and "\<And> x. x \<in> X \<Longrightarrow> s O G x \<subseteq> G x O r"
    shows "s \<subseteq> \<Union> (prot G B X)"
  apply(subgoal_tac "s \<union> r \<in> prot G B X")
   apply blast
  apply(rule prot_intro)
  apply (meson assms(1) assms(2) prot_elim1 wp_ir_act_un)
  by (meson assms(1) assms(3) prot_elim2 wp_ir_act_un2 wp_lipton_gen)

lemma prot_relcomp_compositional:
  assumes "\<And> b. b \<in> B \<Longrightarrow> b \<subseteq> \<langle>s\<rangle> b"
      and "\<And> x. x \<in> X \<Longrightarrow> s O G x \<subseteq> G x O (\<Union> (prot G B X))"
    shows "s \<subseteq> \<Union> (prot G B X)"
  apply(rule prot_relcomp_compositional1[OF prot_join_closed[where rs="prot G B X"]])
    apply blast
  using assms(1) apply blast
  using assms(2) by simp


lemma wk_prot_int_sup_distrib:
  shows "\<Union> (prot G (B \<union> C) (X \<union> Y)) \<subseteq> \<Union> (prot G B X) \<inter> \<Union> (prot G C Y)"
  using Union_Int_subset prot_int_sup_distrib
  by metis


(* A variation on prot follows. But this variation fails
  the prot'_composition lemma below, which I believe is essential. *)

definition "extend G X = {(p, \<langle>G x\<rangle> p) | p x. x \<in> X}"
definition "invars G B X = ((extend G X)\<^sup>*)``B"

lemma invars_induct:
  assumes "p \<in> invars G B X"
      and "\<And> b. b \<in> B \<Longrightarrow> P b"
      and "\<And> p x. p \<in> invars G B X \<Longrightarrow> P p \<Longrightarrow> x \<in> X \<Longrightarrow> P (\<langle>G x\<rangle> p)"
    shows "P p"
  using assms(1) apply(simp add: invars_def)
  using extend_def rtrancl_induct[where r="extend G X"]
  by (smt CollectD ImageE ImageI Pair_inject assms(2) assms(3) invars_def)


definition "prot2 G B X = {r . \<forall> p \<in> invars G B X. p \<subseteq> \<langle>r\<rangle> p}"


lemma prot2_intro_minimal:
  assumes "\<And> p.  p \<in> invars G B X \<Longrightarrow> p \<subseteq> \<langle>r\<rangle> p"
  shows "r \<in> prot2 G B X"
  apply(simp add: prot2_def)
  using assms by blast

lemma prot2_intro2:
  assumes "\<And> b. b \<in> B \<Longrightarrow> b \<subseteq> \<langle>r\<rangle> b"
    and  "\<And> p x s.  x \<in> X \<and> p \<subseteq> \<langle>s\<rangle> p \<and> r \<subseteq> s \<Longrightarrow> \<langle>G x\<rangle> p \<subseteq> \<langle>r\<rangle> (\<langle>G x\<rangle> p)"
  shows "r \<in> prot2 G B X"
  apply(simp add: prot2_def)apply rule
  apply(erule invars_induct)
  using assms(1) apply blast
  using assms(2)
  by blast

lemma prot2_elim1:
  assumes "r \<in> prot2 G B X"
    and  "b \<in> B"
  shows "b \<subseteq> \<langle>r\<rangle> b"
  using assms by(fastforce simp add: prot2_def invars_def)

lemma prot2_elim2:
  assumes "r \<in> prot2 G B X"
      and "p \<in> invars G B X"
    shows "p \<subseteq> \<langle>r\<rangle> p"
  using assms by(simp add: prot2_def invars_def)


lemma prot2_join_closed:
  assumes "rs \<subseteq> prot2 G B X"
  shows "\<Union> rs \<in> prot2 G B X"
  apply(rule prot2_intro_minimal)
  by (meson assms prot2_elim2 subsetD wp_Sup2)

lemma prot'_compositional:
  assumes "r \<in> prot' G B X"
    and "r \<in> prot' G C Y"
  shows "r \<subseteq> \<Union> (prot' G (B \<union> C) (X \<union> Y))"
  oops

end
