theory StoreBuffering
  imports COComp
begin

locale SBFences =  RWActions + TSOFences G Init \<gamma> interf var rval wval F


context SBFences 
begin

datatype PC = L1 | L2 | L3 | End

definition "T1 = (1 :: nat)"
definition "T2 = (2 :: nat)"

lemma T_neq [simp]: "T1 = T2 \<longleftrightarrow> False"
  by(auto simp add: T1_def T2_def)

definition "trd_var x y t \<equiv> if (t = T1) then x else y"
definition "twr_var x y t \<equiv> if (t = T1) then x else y"


fun store_buffering' :: " R \<Rightarrow> R \<Rightarrow> L \<Rightarrow> L \<Rightarrow> nat \<Rightarrow> PC \<Rightarrow> PC AComm"
  where
  "store_buffering' r f x y t L1 = storec False (twr_var x y t) 1 L2" |
  "store_buffering' r f x y t L2 = fencec f 1 L3"  |
  "store_buffering' r f x y t L3 = loada False r (trd_var y x t) f 2 End"  |
  "store_buffering' r f x y t End = skip"

(*at L2, need to say that if T2 reads then defval*)
(*synced on TSO means I have the vmax element in my buffer*)

definition outline2 :: "R \<Rightarrow> R \<Rightarrow> L \<Rightarrow> L \<Rightarrow> T \<Rightarrow> PC \<Rightarrow> (('b, PC) full_state) set"
  where
"outline2 r f x y t pc \<equiv> 
      case pc of
   L1 \<Rightarrow>  (if t = T1 then 
              (Reg T2 (reg f 0) \<union> Reg T2 (reg r 1) \<union> lift\<^sub>1(synced T1 y 1)) 
                  \<inter> Reg T1 (reg f 0)  \<inter> lift\<^sub>1(synced T1 x 0)
          else (if t = T2 then 
              (Reg T1 (reg f 0) \<union> Reg T1 (reg r 1) \<union> lift\<^sub>1(synced T2 x 1)) 
                  \<inter> Reg T2 (reg f 0) \<inter> lift\<^sub>1(synced T2 y 0) 
          else {})) |
   L2 \<Rightarrow>  (if t = T1 then 
              (Reg T2 (reg f 0) \<union> Reg T2 (reg r 1) \<union> lift\<^sub>1(synced T1 y 1)) 
                 \<inter> Reg T1 (reg f 0) \<inter> lift\<^sub>1(synced T1 x 1)
          else (if t = T2 then 
              (Reg T1 (reg f 0) \<union> Reg T1 (reg r 1) \<union> lift\<^sub>1(synced T2 x 1)) 
                 \<inter> Reg T2 (reg f 0) \<inter> lift\<^sub>1(synced T2 y 1) 
          else {})) |
   L3 \<Rightarrow>  (if t = T1 then 
              (Reg T2 (reg f 0) \<union> Reg T2 (reg r 1) \<union> lift\<^sub>1(synced T1 y 1)) 
                 \<inter> Reg T1 (reg f 1) \<inter> lift\<^sub>1(synced T2 x 1)
          else (if t = T2 then 
              (Reg T1 (reg f 0) \<union> Reg T1 (reg r 1) \<union> lift\<^sub>1(synced T2 x 1)) 
                 \<inter> Reg T2 (reg f 1) \<inter> lift\<^sub>1(synced T1 y 1) 
          else {})) |
   End \<Rightarrow> (if t = T1 then 
              (Reg T2 (reg f 0) \<union> Reg T2 (reg f 1) \<union> Reg T2 (reg r 1) \<union> Reg T1 (reg r 1)) 
                  \<inter> Reg T1 (reg f 2) \<inter> lift\<^sub>1(synced T2 x 1)
           else (if t = T2 then 
              (Reg T1 (reg f 0) \<union> Reg T1 (reg f 1) \<union> Reg T1 (reg r 1) \<union> Reg T2 (reg r 1)) 
                  \<inter> Reg T2 (reg f 2) \<inter> lift\<^sub>1(synced T1 y 1) 
           else {}))"


lemma outline2_arb_pc:
  "outline2 r f x y t pc \<subseteq> \<langle>arb_pc t'\<rangle> (outline2 r f x y t pc)"
  apply(cases pc)
  apply (simp_all add: Reg_def lift\<^sub>1_def lift\<^sub>2_def lift_t_def arb_pc_def wp_ext outline2_def synced_def)
     apply safe
  apply simp_all
  using T_neq apply auto
  apply (metis (mono_tags, lifting) Suc_neq_Zero neq0_conv fst_conv less_numeral_extra(3))+
  done

(* We need to move this to CoCOmp *)

(* We need to move this to CoCOmp *)


(* We need to move this to CoCOmp 
lift\<^sub>1 (synced T2 x v)
       \<subseteq> \<langle>full_step {T1, T2}
           (store_buffering' r ra x y)
           (T2, Some (fence ra v))\<rangle>
           (lift\<^sub>1 (synced T2 x v))
*)


lemma fence_sync_transfer: 
  shows "lift\<^sub>1 (synced t x v)
       \<subseteq> \<langle>full_step Ts Pr 
           (t, Some (fence_op r v))\<rangle>
           (lift\<^sub>1 (synced t' x v))"
  apply(rule object_act_lifted_wp)
  using sync_fencing by(auto simp add: mem_act_def)


lemma Reg_mono:
  assumes "B \<subseteq> C"
  shows "Reg t B \<subseteq> Reg t C"
  using assms by(auto simp add: Reg_def lift\<^sub>2_def lift_t_def)
(*
lemma synced_aux_reg:
       "lift\<^sub>1 (synced t x v)
       \<subseteq> \<langle>full_step Ts Pr
           (t, Some (load_aux b x r v ra va))\<rangle>
           (Reg t {reg. reg r = v})"
  sledgehammer
*)
(*
       t' = T2 \<Longrightarrow>
       T2 \<noteq> T1 \<Longrightarrow>
       \<not> b \<and>
       r = ra \<and> trd_var y x T2 = xa \<and> f = raa \<and> va = 2 \<and> End = pc'' \<Longrightarrow>
       t = T1 \<Longrightarrow>
       pc = End \<Longrightarrow>
       pc' = L3 \<Longrightarrow>
       lift\<^sub>1 (synced T2 x (Suc 0))
       \<subseteq> \<langle>full_step {T1, T2} (store_buffering' ra raa x y)
           (T2, Some (load_aux False xa ra v raa 2))\<rangle>
           (Reg T2 {reg. reg ra = Suc 0})
*)

lemma local_correctness_store_buffering:
  assumes "r \<noteq> f"
  assumes "x \<noteq> y"
  shows "local_correctness {T1, T2} (store_buffering' r f x y) (outline2 r f x y) t"
  apply(rule local_correctness_cases)
  using outline2_arb_pc apply fastforce
     apply(case_tac pc; simp)
  apply(case_tac pc; simp)
   apply(case_tac pc; simp)
    apply(simp add: outline2_def) 
  apply (cases t; simp_all)
     apply (intro conjI impI)
  using T_neq apply fastforce
  using SBFences.T2_def SBFences_axioms apply fastforce
  using T1_def apply fastforce
  apply (cases t; simp_all) 
  apply (metis (no_types, lifting) PC.exhaust PC.simps(13) PC.simps(14) PC.simps(15) PC.simps(16) empty_subsetI)
  apply (intro conjI impI)
  using T_neq apply fastforce
     apply(case_tac pc; simp)
  apply (rule wp_ir_pr_conj_intro, rule wp_ir_pr_conj_intro)
  apply (rule wp_ir_pr_conj_elim1, rule wp_ir_pr_conj_elim1)
  apply (rule wp_ir_pr_disj_elim, rule wp_ir_pr_disj_elim)
  apply (rule wp_ir_pr_disj_intro1, rule wp_ir_pr_disj_intro1)
  apply (simp add: act_wp_Reg_stable_other)
  apply (simp add: act_wp_Reg_stable_other wp_ir_pr_disj_intro1 wp_ir_pr_disj_intro2)
         apply (rule wp_ir_pr_disj_intro2)
  apply (meson RWVars.synced_value_\<beta>_stable RWVars_axioms W_wval interf_wp object_write_lifted_wp synced_value_stable)
  apply (simp add: Int_commute Int_left_commute T_neq store_wp_Reg_stable wp_ir_pr_conj_elim1)
  apply (simp add: object_write_lifted_wp synced_value_self_write twr_var_def wp_ir_pr_conj_elim2)      

      apply(case_tac pc; simp)
  apply (rule wp_ir_pr_conj_intro, rule wp_ir_pr_conj_intro)
  apply (rule wp_ir_pr_conj_elim1, rule wp_ir_pr_conj_elim1)
  apply (rule wp_ir_pr_disj_elim, rule wp_ir_pr_disj_elim)
  apply (rule wp_ir_pr_disj_intro1, rule wp_ir_pr_disj_intro1)
  apply (simp add: act_wp_Reg_stable_other)
  apply (simp add: act_wp_Reg_stable_other wp_ir_pr_disj_intro1 wp_ir_pr_disj_intro2)
         apply (rule wp_ir_pr_disj_intro2)
  apply (meson RWVars.synced_value_\<beta>_stable RWVars_axioms W_wval interf_wp object_write_lifted_wp synced_value_stable)
  apply (simp add: Int_commute Int_left_commute T_neq store_wp_Reg_stable wp_ir_pr_conj_elim1)
  apply (simp add: object_write_lifted_wp synced_value_self_write twr_var_def wp_ir_pr_conj_elim2)      

  apply(case_tac pc; simp)
  apply(simp add: outline2_def)
    apply(case_tac pc; simp) apply (intro impI conjI)
      apply(rule wp_ir_pr_conj_intro)+
  using T_neq apply fastforce
  using T_neq apply fastforce
  using T_neq apply fastforce
      apply(rule wp_ir_pr_conj_intro)+
  apply (simp add: Un_subset_iff act_wp_Reg_stable_other fence_sync_transfer le_infI1 wp_ir_pr_disj_intro1 wp_ir_pr_disj_intro2)
  using fence_wp_Reg apply blast
  apply (meson fence_sync_transfer inf.coboundedI2)
      apply(rule wp_ir_pr_conj_intro)+
  apply (simp add: Un_subset_iff act_wp_Reg_stable_other fence_sync_transfer le_infI1 wp_ir_pr_disj_intro1 wp_ir_pr_disj_intro2)
  using fence_wp_Reg apply blast
    apply (meson fence_sync_transfer inf.coboundedI2)
   apply(case_tac pc; simp) 

  apply(cases "t = T1")
    apply(case_tac pc; simp)
  apply(simp add: outline2_def)
    apply(case_tac pc; simp)
      apply(rule wp_ir_pr_conj_intro)+
  apply (simp add: trd_var_def)
      apply(rule wp_ir_pr_conj_elim1)
      apply(rule wp_ir_pr_conj_elim1)
      apply(rule wp_ir_pr_disj_elim)
  apply (meson RWVarsActions.act_wp_Reg_stable_other RWVarsActions_axioms T_neq basic_wp_pres_disj wp_ir_pr_disj_intro1)
      apply(rule wp_ir_pr_disj_intro2)
  apply (metis assms(1) load_aux_wp_Regb load_aux_wp_synced_value)
  apply (simp add: load_aux_wp_Reg)
  apply (meson inf.cobounded2 load_aux_wp_\<beta>_stable subset_trans synced_value_\<beta>_stable)




  apply(simp add: outline2_def)
    apply(case_tac pc; simp add: trd_var_def) apply (intro impI conjI)
      apply(rule wp_ir_pr_conj_intro)+
      apply(rule wp_ir_pr_conj_elim1)+
      apply(rule wp_ir_pr_disj_elim)+
      apply (simp add: act_wp_Reg_stable_other wp_ir_pr_disj_intro1)
  apply (metis act_wp_Reg_stable_other sup_commute wp_ir_pr_disj_intro1)
  apply (metis assms(1) load_aux_wp_Regb load_aux_wp_synced_value sup_commute wp_ir_pr_disj_intro1)
  apply (simp add: load_aux_wp_Reg)
  by (meson inf.cobounded2 load_aux_wp_\<beta>_stable subset_trans synced_value_\<beta>_stable)


lemma noninterference_store_buffering:
  assumes "f \<noteq> r" 
      and "x \<noteq> y"
  shows  "noninterference {T1, T2} (store_buffering' r f x y) (outline2 r f x y) t"
  apply(rule noninterference_cases)
  using outline2_arb_pc apply fastforce
     apply(simp add: outline2_def)
     apply(case_tac "t=T1")
     apply(case_tac pc; case_tac pc'; simp)
     apply(case_tac pc; case_tac pc'; simp)
     apply(simp add: outline2_def)
     apply(case_tac "t=T1")
     apply(case_tac pc; case_tac pc'; simp)
         apply(case_tac pc; case_tac pc'; simp add: twr_var_def; (rule wp_ir_pr_conj_intro)+)
  apply (simp_all add: twr_var_def)
  apply (smt RWVarsActions.store_wp_Reg_stable RWVarsActions_axioms inf.absorb_iff2 inf_assoc inf_commute wp_ir_pr_disj_intro1)
  apply (meson dual_order.trans inf.coboundedI1 inf_le2 store_wp_Reg_stable)
  apply (rule basic_wp_pres)
  apply (metis W_var assms(2) interf_wp2 object_write_lifted_wp synced_value_\<beta>_stable synced_value_stable twr_var_def)
  apply blast
         apply(rule wp_ir_pr_conj_intro)+
  apply (rule subset_trans[where B="(Reg T2 {reg. reg f = 0} \<union> Reg T2 {reg. reg r = 1}) \<union>
        (lift\<^sub>1 (synced T1 y 1) \<inter> lift\<^sub>1 (synced T2 y 0))"],blast)
         apply(rule wp_ir_pr_disj_elim)+
  apply (simp add: store_wp_Reg_stable wp_ir_pr_disj_intro1)
  apply (simp add: store_wp_Reg_stable wp_ir_pr_disj_intro1 wp_ir_pr_disj_intro2)
  apply (rule subset_trans[where B="{}"])
  using synced_diff_threads_vals  
  apply (metis One_nat_def n_not_Suc_n)
  apply blast
  apply (meson dual_order.trans inf.coboundedI1 inf_le2 store_wp_Reg_stable)
  apply (rule basic_wp_pres)
  apply (metis W_var assms(2) interf_wp2 object_write_lifted_wp synced_value_\<beta>_stable synced_value_stable)
  apply blast
  apply(rule wp_ir_pr_conj_intro)+  
  apply (smt RWVarsActions.store_wp_Reg_stable RWVarsActions_axioms inf_left_commute inf_sup_absorb le_infI1 wp_ir_pr_disj_intro1)
  apply (meson dual_order.trans inf.coboundedI1 inf_le2 store_wp_Reg_stable)
  apply (rule basic_wp_pres)
  apply (simp add: interf_wp2 object_write_lifted_wp synced_value_\<beta>_stable synced_value_stable)
  apply blast
  apply(rule wp_ir_pr_conj_intro)+  
  apply (smt RWVarsActions.store_wp_Reg_stable RWVarsActions_axioms inf_left_commute inf_sup_absorb le_infI1 wp_ir_pr_disj_intro1)
  apply (meson dual_order.trans inf.coboundedI1 inf_le2 store_wp_Reg_stable)
  apply (rule basic_wp_pres)
  apply (simp add: interf_wp2 object_write_lifted_wp synced_value_\<beta>_stable synced_value_stable)
      apply blast
  apply (intro impI conjI)
  apply(case_tac pc; case_tac pc'; simp add: twr_var_def; (rule wp_ir_pr_conj_intro)+)
  apply(case_tac pc; case_tac pc'; simp add: twr_var_def; (rule wp_ir_pr_conj_intro)+)
  apply(case_tac pc; case_tac pc'; simp add: twr_var_def; (rule wp_ir_pr_conj_intro)+)
  apply (smt RWVarsActions.store_wp_Reg_stable RWVarsActions_axioms inf_assoc inf_left_commute le_infI1 wp_ir_pr_disj_intro1)
  apply (meson dual_order.trans inf.coboundedI1 inf_le2 store_wp_Reg_stable)
  apply (rule basic_wp_pres)
  apply (metis W_var assms(2) interf_wp2 object_write_lifted_wp synced_value_\<beta>_stable synced_value_stable)
  apply blast
  apply (smt RWVarsActions.store_wp_Reg_stable RWVarsActions_axioms inf_assoc inf_left_commute le_infI1 wp_ir_pr_disj_intro1)
  apply (meson inf.cobounded2 le_infI1 store_wp_Reg_stable subset_trans)
  apply (rule basic_wp_pres)
  apply (metis W_var assms(2) interf_wp2 object_write_lifted_wp synced_value_\<beta>_stable synced_value_stable)
  apply blast
  apply (smt RWVarsActions.store_wp_Reg_stable RWVarsActions_axioms inf_left_commute inf_sup_absorb le_infI1 wp_ir_pr_disj_intro1)
  apply (meson dual_order.trans inf.coboundedI1 inf_le2 store_wp_Reg_stable)
  apply (rule basic_wp_pres)
  apply (simp add: interf_wp2 object_write_lifted_wp synced_value_\<beta>_stable synced_value_stable)
  apply blast
  apply (smt T2_def Un_assoc inf.cobounded1 inf_assoc inf_left_commute store_wp_Reg_stable subset_trans wp_ir_pr_disj_intro1)
  apply (meson inf.cobounded2 le_infI1 store_wp_Reg_stable subset_trans)
  apply (rule basic_wp_pres)
  apply (simp add: interf_wp2 object_write_lifted_wp synced_value_\<beta>_stable synced_value_stable)
  apply blast
  apply (smt PC.case_distrib inf.absorb_iff2 inf_bot_right inf_commute)
  apply simp
  apply simp
    apply(simp add: outline2_def)
    apply(case_tac "t=T1")
     apply (intro conjI impI; simp_all)
     apply(case_tac pc; case_tac pc'; simp ; (rule wp_ir_pr_conj_intro)+)
  apply (meson SBFences.fence_sync_transfer SBFences_axioms le_infI2 wp_ir_pr_disj_intro2)
  using act_wp_Reg_stable_other apply blast
  defer
  apply (meson SBFences.fence_sync_transfer SBFences_axioms le_infI2 wp_ir_pr_disj_intro2)
  using act_wp_Reg_stable_other apply blast
  defer
  apply (meson SBFences.fence_sync_transfer SBFences_axioms le_infI2 wp_ir_pr_disj_intro2)
  using act_wp_Reg_stable_other apply blast
  using fence_sync_transfer apply blast
  apply (simp add: Un_assoc Un_left_commute fence_wp_Reg wp_ir_pr_disj_intro1)
  using act_wp_Reg_stable_other apply blast
  apply (simp add: fence_sync_transfer inf.coboundedI1 le_infI2)
    apply(case_tac "t=T1")
  apply(case_tac pc; case_tac pc'; simp)
  apply(case_tac pc; case_tac pc'; simp)
     apply (intro conjI impI; simp_all)
  apply(rule wp_ir_pr_conj_intro)+  
  apply (meson SBFences.fence_sync_transfer SBFences_axioms le_infI2 wp_ir_pr_disj_intro2)
  using T_neq act_wp_Reg_stable_other apply blast
  defer
     apply (intro conjI impI; simp_all)
  apply(rule wp_ir_pr_conj_intro)+  
  apply (meson SBFences.fence_sync_transfer SBFences_axioms le_infI2 wp_ir_pr_disj_intro2)
  using T_neq act_wp_Reg_stable_other apply blast
         defer
     apply (intro conjI impI; simp_all)
  apply(rule wp_ir_pr_conj_intro)+  
  apply (meson SBFences.fence_sync_transfer SBFences_axioms le_infI2 wp_ir_pr_disj_intro2)        
  using T_neq act_wp_Reg_stable_other apply blast
  using fence_sync_transfer apply blast
     apply (intro conjI impI; simp_all)
  apply(rule wp_ir_pr_conj_intro)+  
  apply (metis fence_wp_Reg sup_commute wp_ir_pr_disj_intro1)
  using T_neq act_wp_Reg_stable_other apply blast
  apply (simp add: fence_sync_transfer inf.coboundedI1 le_infI2)
    apply(simp add: outline2_def)
    apply(case_tac "t=T1")
  apply(case_tac pc; case_tac pc'; simp)
  apply(case_tac pc; case_tac pc'; simp)
    apply(simp add: outline2_def)
    apply(case_tac "t=T1")
  apply(case_tac pc; case_tac pc'; simp)
  apply(rule wp_ir_pr_conj_intro)+  
  apply (simp add: le_infI2 load_aux_wp_\<beta>_stable synced_value_\<beta>_stable wp_ir_pr_disj_intro2)
  using act_wp_Reg_stable_other apply blast
  apply (simp add: le_infI2 load_aux_wp_\<beta>_stable synced_value_\<beta>_stable wp_ir_pr_conj_elim3)
  apply(rule wp_ir_pr_conj_intro)+  
  apply (simp add: le_infI2 load_aux_wp_\<beta>_stable synced_value_\<beta>_stable wp_ir_pr_disj_intro2)
  using act_wp_Reg_stable_other apply blast  
  apply (simp add: le_infI2 load_aux_wp_\<beta>_stable synced_value_\<beta>_stable wp_ir_pr_conj_elim3)
  apply(rule wp_ir_pr_conj_intro)+  
  apply (simp add: le_infI2 load_aux_wp_\<beta>_stable synced_value_\<beta>_stable wp_ir_pr_disj_intro2)
  using act_wp_Reg_stable_other apply blast
  apply (simp add: le_infI2 load_aux_wp_\<beta>_stable synced_value_\<beta>_stable wp_ir_pr_conj_elim3)
       apply(rule wp_ir_pr_conj_intro)+  
  apply (rule subset_trans[where B="lift\<^sub>1 (synced T2 x (Suc 0))"],blast)
       apply(rule wp_ir_pr_disj_intro1, rule wp_ir_pr_disj_intro2)  
  apply (case_tac "v=1")
  apply (simp add: load_aux_wp_synced_value trd_var_def)
  using assms(1) load_aux_wp_Regb apply blast
  apply (simp add: load_aux_wp_synced_value trd_var_def)
  using act_wp_Reg_stable_other apply blast
  apply (simp add: le_infI2 load_aux_wp_\<beta>_stable synced_value_\<beta>_stable wp_ir_pr_conj_elim3)
    apply(case_tac "t=T1")
  apply(case_tac pc; case_tac pc'; simp)
  apply(case_tac pc; case_tac pc'; simp)
     apply (intro conjI impI; simp_all)
  apply(rule wp_ir_pr_conj_intro)+  
  apply (simp add: le_infI2 load_aux_wp_\<beta>_stable synced_value_\<beta>_stable wp_ir_pr_disj_intro2)
  using T_neq act_wp_Reg_stable_other apply blast
  apply (simp add: le_infI2 load_aux_wp_\<beta>_stable synced_value_\<beta>_stable wp_ir_pr_conj_elim3)
     apply (intro conjI impI; simp_all)
  apply(rule wp_ir_pr_conj_intro)+  
  apply (simp add: le_infI2 load_aux_wp_\<beta>_stable synced_value_\<beta>_stable wp_ir_pr_disj_intro2)
  using T_neq act_wp_Reg_stable_other apply blast
  apply (simp add: le_infI2 load_aux_wp_\<beta>_stable synced_value_\<beta>_stable wp_ir_pr_conj_elim3)
     apply (intro conjI impI; simp_all)
  apply(rule wp_ir_pr_conj_intro)+  
  apply (simp add: le_infI2 load_aux_wp_\<beta>_stable synced_value_\<beta>_stable wp_ir_pr_disj_intro2)
  using T_neq act_wp_Reg_stable_other apply blast
  apply (simp add: le_infI2 load_aux_wp_\<beta>_stable synced_value_\<beta>_stable wp_ir_pr_conj_elim3)
     apply (intro conjI impI; simp_all add: trd_var_def)
      apply(rule wp_ir_pr_conj_intro)+  
  apply (rule subset_trans[where B="lift\<^sub>1 (synced T1 y (Suc 0))"],blast)
       apply(rule wp_ir_pr_disj_intro1, rule wp_ir_pr_disj_intro2)  
  apply (metis assms(1) load_aux_wp_Regb load_aux_wp_synced_value)
  using T_neq act_wp_Reg_stable_other apply blast
  apply (simp add: le_infI2 load_aux_wp_\<beta>_stable synced_value_\<beta>_stable wp_ir_pr_conj_elim3)
  apply (smt F_wval RWVars.read_preserves_all RWVarsActions.mem_op.case(3) RWVarsActions.object_act_lifted_wp RWVarsActions_axioms RWVars_axioms inf.absorb_iff2 inf.cobounded2 le_infI1 mem_act_def synced_value_\<beta>_stable)
  apply (smt F_wval RWVars.read_preserves_all RWVarsActions.mem_op.case(3) RWVarsActions.object_act_lifted_wp RWVarsActions_axioms RWVars_axioms inf.absorb_iff2 inf.cobounded2 le_infI1 mem_act_def synced_value_\<beta>_stable)
  apply (smt F_wval RWVars.read_preserves_all RWVarsActions.mem_op.case(3) RWVarsActions.object_act_lifted_wp RWVarsActions_axioms RWVars_axioms inf.absorb_iff2 inf.cobounded2 le_infI1 mem_act_def synced_value_\<beta>_stable)
  by (smt F_wval RWVars.read_preserves_all RWVarsActions.mem_op.case(3) RWVarsActions.object_act_lifted_wp RWVarsActions_axioms RWVars_axioms inf.absorb_iff2 inf.cobounded2 le_infI1 mem_act_def synced_value_\<beta>_stable)


definition "lb_init r f x y \<equiv> lift\<^sub>1 (synced T1 x 0) \<inter>  lift\<^sub>1 (synced T2 x 0) \<inter> 
                                lift\<^sub>1 (synced T1 y 0) \<inter>  lift\<^sub>1 (synced T2 y 0)
                                     \<inter> Reg T1 (reg r 0) \<inter> Reg T2 (reg r 0) 
                                     \<inter> Reg T1 (reg f 0) \<inter> Reg T2 (reg f 0) \<inter>
  {s . thr_pc s T1 = L1} \<inter> {s . thr_pc s T2 = L1}"

theorem
  assumes "x \<noteq> y"
  and "f \<noteq> r"
  shows "reach (\<Union> e. full_step {T1, T2} (store_buffering' r f x y) e) (lb_init r f x y) \<subseteq> inv_of {T1, T2} (outline2 r f x y)"
  apply(rule owicki_gries[OF local_correctness_store_buffering noninterference_store_buffering])
  using assms apply blast
  using assms  apply blast
  using assms  apply blast
  using assms  apply blast
  apply (simp add: lb_init_def outline2_def)
  apply rule
  apply rule
  apply rule
  using T_neq apply presburger
  apply rule
  done

end
