theory StoreBuffering2
  imports COComp
begin

locale SBFences =  RWActions + TSOFences G Init \<gamma> interf var rval wval F


context SBFences 
begin

datatype PC = L1 | L3 | End

definition "T1 = (1 :: nat)"
definition "T2 = (2 :: nat)"

lemma T_neq [simp]: "T1 = T2 \<longleftrightarrow> False"
  by(auto simp add: T1_def T2_def)

definition "trd_var x y t \<equiv> if (t = T1) then x else y"
definition "twr_var x y t \<equiv> if (t = T1) then x else y"


fun store_buffering' :: " R \<Rightarrow> L \<Rightarrow> L \<Rightarrow> nat \<Rightarrow> PC \<Rightarrow> PC AComm"
  where
  "store_buffering' r x y t L1 = storec False (twr_var x y t) 1 L3" |
  "store_buffering' r x y t L3 = loadc False r (trd_var y x t) End"  |
  "store_buffering' r x y t End = skip"

definition outline2 :: "R \<Rightarrow> L \<Rightarrow> L \<Rightarrow> T \<Rightarrow> PC \<Rightarrow> (('b, PC) full_state) set"
  where
"outline2 r x y t pc \<equiv> 
      case pc of
   L1 \<Rightarrow>  (if t = T1 then 
              lift\<^sub>1(constrained_value t y {0,1})
          else (if t = T2 then 
              lift\<^sub>1(constrained_value t x {0,1})
          else {})) |
   L3 \<Rightarrow>  (if t = T1 then 
              lift\<^sub>1(constrained_value t y {0,1})
          else (if t = T2 then 
              lift\<^sub>1(constrained_value t x {0,1})
          else {})) |
   End \<Rightarrow> (if t = T1 then 
              Reg T1 (reg r 0) \<union> Reg T1 (reg r 1)
           else (if t = T2 then 
              Reg T2 (reg r 0) \<union> Reg T2 (reg r 1)
           else {}))"


lemma outline2_arb_pc:
  "outline2 r x y t pc \<subseteq> \<langle>arb_pc t'\<rangle> (outline2 r x y t pc)"
  apply(cases pc)
  apply (simp_all add: Reg_def lift\<^sub>1_def lift\<^sub>2_def lift_t_def arb_pc_def wp_ext outline2_def synced_def)
     apply safe
  apply simp_all
  using T_neq apply auto
                      apply (metis (mono_tags, lifting) Suc_neq_Zero neq0_conv fst_conv less_numeral_extra(3))+
  done

lemma fence_sync_transfer: 
  shows "lift\<^sub>1 (synced t x v)
       \<subseteq> \<langle>full_step Ts Pr 
           (t, Some (fence_op r v))\<rangle>
           (lift\<^sub>1 (synced t' x v))"
  apply(rule object_act_lifted_wp)
  using sync_fencing by(auto simp add: mem_act_def)


lemma Reg_mono:
  assumes "B \<subseteq> C"
  shows "Reg t B \<subseteq> Reg t C"
  using assms by(auto simp add: Reg_def lift\<^sub>2_def lift_t_def)



lemma constrained_value_stable_diff_var:
  assumes "mem_act ao a"
  and "var a \<noteq> x"
  shows "lift\<^sub>1 (constrained_value t x v) 
      \<subseteq> \<langle>full_step Ts Pr (t', Some ao)\<rangle>
           (lift\<^sub>1 (constrained_value t x v))"
  using assms
   apply(case_tac "ao")
  apply(simp add: mem_act_def)
  using G_\<beta>_stable
  apply (simp add: constrained_value_\<beta>_stable load_wp_\<beta>_stable)
  using G_\<beta>_stable
   apply(simp add: mem_act_def)
  apply (simp add: constrained_value_\<beta>_stable constrained_value_stable interf_wp object_write_lifted_wp)
  apply (metis (mono_tags, lifting) RWVars.constrained_value_\<beta>_stable RWVars.constrained_value_stable RWVars_axioms interf_wp2 mem_act_def mem_op.simps(19) object_act_lifted_wp)
  by (meson RWVars.constrained_value_\<beta>_stable RWVars_axioms load_aux_wp_\<beta>_stable)


lemma constrained_value_stable_load:
  shows "lift\<^sub>1 (constrained_value t x v) 
      \<subseteq> \<langle>full_step Ts Pr (t', Some (load b y r u))\<rangle>
           (lift\<^sub>1 (constrained_value t x v))"
  by (simp add: constrained_value_\<beta>_stable load_wp_\<beta>_stable)



lemma constrained_value_stable_store_diff_var:
  assumes "y \<noteq> x"
  shows "lift\<^sub>1 (constrained_value t x v) 
      \<subseteq> \<langle>full_step Ts Pr (t', Some (store b y u))\<rangle>
           (lift\<^sub>1 (constrained_value t x v))"
  using constrained_value_stable_diff_var  
  by (simp add: constrained_value_stable_diff_var assms mem_act_def)


lemma constrained_value_stable_store_existing:
  assumes "u \<in> v"
  shows "lift\<^sub>1 (constrained_value t x v) 
      \<subseteq> \<langle>full_step Ts Pr (t', Some (store b x u))\<rangle>
           (lift\<^sub>1 (constrained_value t x v))"
  using constrained_value_stable_diff_var  
  by (smt RWVars.constrained_value_stable RWVarsActions.object_write_lifted_wp RWVarsActions_axioms RWVars_axioms W_wval assms constrained_value_\<beta>_stable image_iff interf_wp)

lemma constrained_value_stable_load_Reg:
  shows "lift\<^sub>1 (constrained_value t x {v, v'}) 
      \<subseteq> \<langle>full_step Ts Pr (t, Some (load b x r u))\<rangle>
           (Reg t {reg. reg r = v} \<union> Reg t {reg. reg r = v'})"
  apply (case_tac "u = v", simp)
  apply (simp add: constrained_value_def lift\<^sub>1_def mem_act_def)
  apply (auto simp add: wp_ext Reg_def full_step_def co_step_def par_step_def
  mem_act_def lift\<^sub>2_def lift_t_def  arb_pc_def)[1]
  apply (case_tac "u = v'", simp)
  apply(rule wp_ir_pr_disj_intro2)
  apply (simp add: constrained_value_def lift\<^sub>1_def mem_act_def)
  apply (auto simp add: wp_ext Reg_def full_step_def co_step_def par_step_def
  mem_act_def lift\<^sub>2_def lift_t_def  arb_pc_def)[1]
  apply (simp add: constrained_value_def lift\<^sub>1_def mem_act_def not_value_def)
  apply (auto simp add: wp_ext Reg_def full_step_def co_step_def par_step_def
  mem_act_def lift\<^sub>2_def lift_t_def  arb_pc_def)[1]
  using R_Rds act_Tr_set by blast

lemma local_correctness_store_buffering:
  assumes "x \<noteq> y"
  shows "local_correctness {T1, T2} (store_buffering' r x y) (outline2 r x y) t"
  apply(rule local_correctness_cases)
  using outline2_arb_pc apply fastforce
     apply(case_tac pc; simp)
  apply(case_tac pc; simp)
   apply(case_tac pc; simp)
    apply(simp add: outline2_def) 
  apply (cases t; simp_all)
     apply (intro conjI impI)
  using T_neq apply fastforce
  using SBFences.T2_def SBFences_axioms apply fastforce
  using T1_def apply fastforce
  apply (cases "t=T1"; simp_all) 

  apply (simp add: constrained_value_stable_load_Reg trd_var_def)
  apply (simp add: constrained_value_stable_load_Reg trd_var_def)
    apply(simp add: outline2_def) 
  apply (cases t; simp_all)
     apply (intro conjI impI)
  using T_neq apply fastforce
     apply(case_tac pc; simp)
  apply (simp add: T2_def)
  using T1_def apply linarith
  apply (metis (no_types, hide_lams) PC.case_distrib empty_subsetI inf_sup_ord(1) subset_empty)
  apply (cases t; simp_all)
     apply (intro conjI impI)
     apply(case_tac pc; simp)
  using T_neq apply auto[1]
     apply(case_tac pc; simp)
    apply(simp add: twr_var_def) 
  apply (metis assms constrained_value_stable_store_diff_var)
     apply(case_tac pc; simp)
  apply (metis assms constrained_value_stable_store_diff_var twr_var_def)
  apply (metis (no_types, lifting) PC.case_distrib inf.absorb_iff2 wp_ir_pr_bot)
    apply(simp add: outline2_def) 
  apply (cases t; simp_all)
     apply (intro conjI impI)
  using T_neq apply fastforce
     apply(case_tac pc; simp)
  apply (simp add: T2_def)
  using T1_def apply linarith
  apply (metis (no_types, hide_lams) PC.case_distrib empty_subsetI inf_sup_ord(1) subset_empty)
  apply (cases t; simp_all)
     apply (intro conjI impI)
     apply(case_tac pc; simp)
  using T_neq apply auto[1]
     apply(case_tac pc; simp)
    apply(simp add: twr_var_def) 
     apply(case_tac pc; simp)
  
  apply (metis (no_types, lifting) PC.case_distrib inf.absorb_iff2 wp_ir_pr_bot)
    apply(simp add: outline2_def) 
     apply(case_tac pc; simp)
    apply(simp add: outline2_def) 
by(case_tac pc; simp)


lemma noninterference_store_buffering:
  assumes "x \<noteq> y"
  shows  "noninterference {T1, T2} (store_buffering' r x y) (outline2 r x y) t"
  apply(rule noninterference_cases)
  using outline2_arb_pc apply fastforce
     apply(simp add: outline2_def)
     apply(case_tac "t=T1")
     apply(case_tac pc; case_tac pc'; simp)
     apply(case_tac pc; case_tac pc'; simp)
  apply (simp add: constrained_value_stable_load le_infI1)
  
        apply (simp add: constrained_value_stable_load le_infI1)
  apply (smt RWVarsActions.act_wp_Reg_stable_other RWVarsActions_axioms Un_subset_iff inf.coboundedI1 subset_trans wp_pred_sup_semi_distrib)
     apply(case_tac pc; case_tac pc'; simp)
  apply (simp add: constrained_value_stable_load le_infI1)
  apply (simp add: constrained_value_stable_load le_infI1)
  apply (smt RWVarsActions.act_wp_Reg_stable_other RWVarsActions_axioms Un_subset_iff dual_order.trans inf.coboundedI1 wp_pred_sup_semi_distrib)
     apply(simp add: outline2_def)
     apply(case_tac "t'",simp_all)
     apply (intro conjI impI)
     apply(case_tac pc; case_tac pc'; simp)
     apply(case_tac pc; case_tac pc'; simp)
  using T2_def apply linarith
  apply (simp add: T2_def)
  using T2_def apply linarith
  apply (simp add: T2_def)
  using T1_def apply linarith
  using T1_def apply linarith
  apply auto[1]
  apply simp
  apply simp
  apply simp
     apply (intro conjI impI)
     apply(case_tac pc; case_tac pc'; simp)
     apply(case_tac pc; case_tac pc'; simp)
              apply (simp add: twr_var_def)
apply(rule wp_ir_pr_conj_elim1)
  apply (simp add: constrained_value_stable_store_existing)
  apply (simp add: constrained_value_stable_store_existing inf.coboundedI1 twr_var_def)
  apply (simp add: le_infI1 store_wp_Reg_stable wp_ir_pr_disj_intro1 wp_ir_pr_disj_intro2)
  apply (smt PC.exhaust PC.simps(7) PC.simps(8) PC.simps(9) inf.coboundedI1 wp_ir_pr_bot)
     apply(case_tac pc; case_tac pc'; simp)
     apply(case_tac pc; case_tac pc'; simp)
  apply (simp add: constrained_value_stable_store_existing inf.coboundedI1 twr_var_def)
  apply (simp add: constrained_value_stable_store_existing inf.coboundedI1 twr_var_def)
  apply (meson Un_mono inf_le1 store_wp_Reg_stable subset_trans wp_pred_sup_semi_distrib)
     apply(case_tac pc; case_tac pc'; simp)
  apply (metis PC.case_distrib empty_subsetI inf.cobounded1 subset_empty)
  apply blast+
     apply(simp add: outline2_def)
     apply(case_tac "t'",simp_all)
     apply (intro conjI impI)
     apply(case_tac pc; case_tac pc'; simp)
     apply(case_tac pc; case_tac pc'; simp)
  using T2_def apply linarith
          apply (simp add: T2_def)
  using T1_def apply linarith
  using T1_def apply linarith
  apply simp_all
     apply(case_tac pc; case_tac pc'; simp)
     apply(simp add: outline2_def)
     apply(case_tac "t'",simp_all)
     apply (intro conjI impI)
     apply(case_tac pc; case_tac pc'; simp)
     apply(case_tac pc; case_tac pc'; simp)
  using T2_def apply linarith
          apply (simp add: T2_def)
  using T1_def apply linarith
  using T1_def apply linarith
  apply simp_all
     apply(case_tac pc; case_tac pc'; simp)
     apply(simp add: outline2_def)
     apply(case_tac "t'",simp_all)
     apply (intro conjI impI)
     apply(case_tac pc; case_tac pc'; simp)
     apply(case_tac pc; case_tac pc'; simp)
  using T2_def apply linarith
          apply (simp add: T2_def)
  using T1_def apply linarith
  using T1_def apply linarith
  apply simp_all
  by (case_tac pc; case_tac pc'; simp)


definition "lb_init r x y \<equiv> lift\<^sub>1
                 (constrained_value T2 x
                   {0, Suc 0}) \<inter>  lift\<^sub>1
                 (constrained_value T1 y
                   {0, Suc 0})
                                     \<inter> {s . thr_pc s T1 = L1} \<inter> {s . thr_pc s T2 = L1}"

theorem
  assumes "x \<noteq> y"
  shows "reach (\<Union> e. full_step {T1, T2} (store_buffering' r x y) e) (lb_init r x y) \<subseteq> inv_of {T1, T2} (outline2 r x y)"
  apply(rule owicki_gries[OF local_correctness_store_buffering noninterference_store_buffering])
  using assms apply blast
  using assms  apply blast
  apply (simp add: lb_init_def outline2_def)
  apply rule
  apply rule
  apply rule
  using T_neq apply presburger
  apply rule
  done

end
