theory TransitionSystems
  imports WP_SP DualModalOps Simulations
begin

lemma recomp_subs_right:
  assumes "\<And> a. R O S a \<subseteq> T"
  shows "R O (\<Union> a. S a) \<subseteq> T"
  using assms by blast

lemma recomp_subs_left:
  assumes "\<And> a. R a O S \<subseteq> T"
  shows "(\<Union> a. R a) O S \<subseteq> T"
  using assms by blast


locale TSys =
  fixes G :: "nat \<times> 'b \<Rightarrow> 's rel"
    and Init :: "'s set"
begin

definition Tr :: "nat \<Rightarrow> 'b set \<Rightarrow> 's rel" (infix "\<triangleright>" 90)
  where
  "Tr t A \<equiv> \<Union> a \<in> A. G (t, a)"


lemma [simp]: "t \<triangleright> {} = {}" by(simp add: Tr_def)
lemma Tr_G: "t \<triangleright> {a} = G (t, a)" by(simp add: Tr_def)

lemma Tr_t_a_ext: "t \<triangleright> \<Union>X = (\<Union> A \<in> X. \<Union> a \<in> A. G (t, a))"
  by(simp add: Tr_def)

lemma Domain_Sup_distrib:
  "Domain ( \<Union> a \<in> A. G (t, a)) = (\<Union> a \<in> A. Domain (G (t, a)))"
  by auto

lemma wp_emp_not_trans:
  "s \<in> \<langle>t \<triangleright> A\<rangle>{} \<Longrightarrow> (s, s') \<in> G (t, b) \<Longrightarrow> b \<notin> A"
  apply(simp add: wp_ext Tr_def)
  by (metis)

lemma wp_set_Int:
  "\<langle>t \<triangleright> A\<rangle> P = (\<Inter> a \<in> A. \<langle>G (t, a)\<rangle> P)"
  by(auto simp add: wp_ext Tr_def)

lemma wp_stable_set_cond:
  assumes "\<And> a. a \<in> X \<Longrightarrow> \<langle>G (t, a)\<rangle> P \<subseteq> \<langle>A\<rangle> (\<langle>G (t, a)\<rangle> P)"
  shows "\<langle>t \<triangleright> X\<rangle> P \<subseteq> \<langle>A\<rangle> (\<langle>t \<triangleright> X\<rangle> P)"
  using assms apply(simp add: wp_set_Int)
  by (simp add: Int_wp_pres2)

lemma wp_stable_set_action:
  assumes "\<And> a. a \<in> X \<Longrightarrow> \<langle>A\<rangle> P \<subseteq> \<langle>G (t, a)\<rangle> (\<langle>A\<rangle> Q)"
  shows "\<langle>A\<rangle> P \<subseteq> \<langle>t \<triangleright> X\<rangle> (\<langle>A\<rangle> Q)"
  using assms apply(simp add: wp_set_Int) by blast

lemma wp_stable_sets:
  assumes "\<And> a b. a \<in> X \<Longrightarrow> b \<in> Y \<Longrightarrow> \<langle>G (t, a)\<rangle> P \<subseteq> \<langle>G (t', b)\<rangle> (\<langle>G (t, a)\<rangle> P)"
  shows "\<langle>t \<triangleright> X\<rangle> P \<subseteq> \<langle>t' \<triangleright> Y\<rangle> (\<langle>t \<triangleright> X\<rangle> P)"
  apply(rule wp_stable_set_cond)
  apply(rule wp_stable_set_action)
  by(rule assms(1); simp)


lemma set_wp_left:
  assumes "\<And> a. a \<in> X \<Longrightarrow> p \<subseteq> \<langle>G (t,a)\<rangle> q"
  shows "p \<subseteq> \<langle>t \<triangleright> X\<rangle> q"
  by (simp add: INT_greatest assms wp_set_Int)

end

definition "trel G B = B \<union> (\<Union> e. G e)"

definition "reach R I = (R\<^sup>*)``I"

lemma reach_intro:
  assumes "I \<subseteq> P"
      and "\<And> e s s'. (s, s') \<in> G e \<Longrightarrow> s \<in> P \<Longrightarrow> s' \<in> P"
      and "\<And> s s'. (s, s') \<in> B \<Longrightarrow> s \<in> P \<Longrightarrow> s' \<in> P"
    shows "reach (trel G B) I \<subseteq> P"
  apply(simp add: trel_def reach_def Image_def) apply clarsimp
  apply(erule rtrancl_induct) using assms apply blast
  using assms by blast

lemma reach_intro1:
  assumes "I \<subseteq> P"
      and "\<And> e s s'. (s, s') \<in> F e \<Longrightarrow> s \<in> P \<Longrightarrow> s' \<in> P"
    shows "reach (\<Union> e. F e) I \<subseteq> P"
  apply(simp add: reach_def Image_def) apply clarsimp
  apply(erule rtrancl_induct) using assms apply blast
  using assms by blast

lemma reach_step:
  "c \<in> reach R I \<Longrightarrow> (c, c') \<in> R \<Longrightarrow> c' \<in> reach R I"
  apply(simp add: reach_def)
  by (metis ImageI r_into_rtrancl relcomp_Image rtrancl_idemp_self_comp)

lemma reach_external:
  "c \<in> reach (trel G B) I \<Longrightarrow> (c, c') \<in> G e \<Longrightarrow> c' \<in> reach (trel G B) I"
  apply(simp add: trel_def) using reach_step
  by (smt UnCI UnionI rangeI)

lemma reach_internal:
  "c \<in> reach (trel G B) I \<Longrightarrow> (c, c') \<in> B \<Longrightarrow> c' \<in> reach (trel G B) I"
  apply(simp add: trel_def) using reach_step
  by (smt UnCI UnionI rangeI)

definition
  "backsim_init R I I' =
    (\<forall> c a. R c a \<and> c \<in> I \<longrightarrow> a \<in> I')"

definition
  "backsim_step R G B I G' =
    (\<forall> c. \<forall> e c' a'. c \<in> reach (trel G B) I \<and> R c' a' \<and> (c, c') \<in> G e \<longrightarrow> (\<exists> a. (a, a') \<in> G' e \<and> R c a))"

definition
  "backsim_int R G B I B' =
    (\<forall> c. \<forall> c' a'. c \<in> reach (trel G B) I \<and> R c' a' \<and> (c, c') \<in> B \<longrightarrow> (\<exists> a. (a, a') \<in> B' \<and> R c a))"

definition "backsim R G B I G' B' I' \<equiv>
   backsim_init R I I' \<and> backsim_step R G B I G' \<and> backsim_int R G B I B'"

lemma backsim_intro:
  assumes "\<And> c a. R c a \<and> c \<in> I \<Longrightarrow> a \<in> I'"
   and "\<And> e c c' a'. c \<in> reach (trel G B) I \<and> R c' a' \<and> (c, c') \<in> G e \<Longrightarrow> \<exists> a. (a, a') \<in> G' e \<and> R c a"
   and "\<And> c c' a'. c \<in> reach (trel G B) I \<and> R c' a' \<and> (c, c') \<in> B \<Longrightarrow> \<exists> a. (a, a') \<in> B' \<and> R c a"
  shows "backsim R G B I G' B' I'"
  using assms by(fastforce simp add: backsim_def backsim_init_def backsim_step_def backsim_int_def)

end
