theory UnifiedAxioms
  imports TransitionSystems
begin



lemma Inf_inf_tool:
  "a \<in> X \<Longrightarrow> (\<Inter> a \<in> X. p a) \<inter> q = (\<Inter> a \<in> X. (p a \<inter> q))"
  by force

lemma Inf_tool:
  "a \<in> X \<Longrightarrow> p a \<subseteq> q \<Longrightarrow> (\<Inter> a \<in> X. p a) \<subseteq> q"
  by blast



abbreviation "wvs G \<gamma> \<equiv> (\<Union> (bndsim G {\<gamma> x | x. True}))"

locale ViewTSys = TSys +
  fixes \<gamma> interf
  assumes interference_valid:
    "G e \<subseteq> wvs G \<gamma> O interf e O wvs G \<gamma>"
  assumes \<gamma>_G_stable:
    "(\<gamma> (t, a)) \<subseteq> \<langle>interf (t, b)\<rangle> (\<gamma> (t, a))"
(*  assumes vmax_\<beta>_confluence:
    "x \<in> \<gamma> (t, a) \<Longrightarrow> (x, y) \<in> G (t, a) \<Longrightarrow> (x, y') \<in> wvs G \<gamma> \<Longrightarrow>
            (\<exists> z. (y, z) \<in> wvs G \<gamma> \<and> (y', z) \<in> G (t, a))" *)
  assumes vmax_inclusion:
    "x \<in> \<gamma> (t', a) \<Longrightarrow> (x, y') \<in> G (t', a) \<Longrightarrow> \<exists> y z. (x, y) \<in> G (t, a) \<and> (y, z) \<in> wvs G \<gamma> \<and> (y', z) \<in> wvs G \<gamma>"
begin



definition "\<beta> = (\<Union> (bndsim G {\<gamma> e | e. True}))"


lemma \<beta>_\<gamma>_bndsim:
  "\<beta> \<in> bndsim G {\<gamma> e | e. True}"
  by (simp add: \<beta>_def bndsim_join_closed)

lemma \<beta>_trans:
   "\<beta> O \<beta> = \<beta>"
  apply rule
  using bndsim_trans[where r=\<beta> and s=\<beta>]
  using \<beta>_\<gamma>_bndsim \<beta>_def apply blast
  using \<beta>_def bndsim_refl by fastforce

lemma \<beta>_refl:
  "Id \<subseteq> \<beta>"
  by (simp add: Sup_upper \<beta>_def bndsim_refl)

lemma interf_wp:
  assumes "p \<subseteq> \<langle>\<beta>\<rangle> p"
      and "p \<subseteq> \<langle>interf e\<rangle> p"
  shows "p \<subseteq> \<langle>G e\<rangle> p"
  using wp_act_antimono2[OF interference_valid] assms
  by (metis (no_types, lifting) \<beta>_def interpolation_left wp_pres_compose)

lemma interf_wp2:
  assumes "p \<subseteq> \<langle>\<beta>\<rangle> p"
  assumes "q \<subseteq> \<langle>\<beta>\<rangle> q"
      and "p \<subseteq> \<langle>interf e\<rangle> q"
  shows "p \<subseteq> \<langle>G e\<rangle> q"
  using wp_act_antimono2[OF interference_valid] assms
  by (metis (no_types, lifting) \<beta>_def interpolation_left wp_pres_compose)

lemma \<beta>_commutes:
  "\<beta> O G e \<subseteq> G e O \<beta>"
  by(rule bndsim_commutes[OF \<beta>_\<gamma>_bndsim])

lemma \<gamma>_\<beta>_stable:
  "\<gamma> x \<subseteq> \<langle>\<beta>\<rangle> (\<gamma> x)"
  by (metis (mono_tags, lifting) IntD1 \<beta>_\<gamma>_bndsim bndsim_def mem_Collect_eq)


lemma G_\<beta>_stable:
  assumes "P \<subseteq> \<langle>\<beta>\<rangle> P"
  shows "\<langle>G e\<rangle> P \<subseteq> \<langle>\<beta>\<rangle> (\<langle>G e\<rangle> P)"
  using \<beta>_commutes
  by (simp add: wp_lipton assms)

lemma set_\<beta>_stable:
  assumes "P \<subseteq> \<langle>\<beta>\<rangle> P"
  shows "\<langle>t \<triangleright> X\<rangle> P \<subseteq> \<langle>\<beta>\<rangle> (\<langle>t \<triangleright> X\<rangle> P)"
  apply(rule wp_stable_set_cond)
  using G_\<beta>_stable[OF assms(1)] by blast


definition "\<Gamma> t A \<equiv> \<Inter> a \<in> A. \<gamma> (t, a)"

lemma \<Gamma>_\<beta>_stable:
  "\<Gamma> t A \<subseteq> \<langle>\<beta>\<rangle> (\<Gamma> t A)"
  apply(simp add: \<Gamma>_def)
  by (simp add: Int_wp_pres2 \<gamma>_\<beta>_stable)

lemma \<Gamma>_self_stable:
  "\<Gamma> t A \<subseteq> \<langle>G (t, a)\<rangle> (\<Gamma> t A)"
  apply(simp add: \<Gamma>_def)
  using \<gamma>_G_stable interf_wp \<Gamma>_\<beta>_stable
  by (simp add: Int_wp_pres2 \<gamma>_\<beta>_stable)

lemma \<Gamma>_antimono:
  "A \<subseteq> B \<Longrightarrow> \<Gamma> t' B \<subseteq> \<Gamma> t' A"
  by(auto simp add: \<Gamma>_def)

lemma \<gamma>_is_\<Gamma>:
  "\<gamma> (t, a) = \<Gamma> t {a}"
  by(simp add: \<Gamma>_def)

lemma \<gamma>_\<Gamma>_lift:
  assumes "\<And> x. x \<in> X \<Longrightarrow> \<gamma> (t, x) \<subseteq> \<langle>A\<rangle> (\<gamma> (t, x))"
  shows "\<Gamma> t X \<subseteq> \<langle>A\<rangle> (\<Gamma> t X)"
  using assms by(auto simp add: \<Gamma>_def wp_ext)

lemma \<beta>_\<Gamma>_bndsim:
  "\<beta> \<in> bndsim G {\<Gamma> t A | t A. True}"
  apply(simp add: bndsim_def)
  apply rule
  using \<Gamma>_\<beta>_stable apply blast
  using \<beta>_\<gamma>_bndsim bndsim_def by auto

lemma \<gamma>_act_dwp_wp_transfer:
  assumes "P \<subseteq> \<langle>\<beta>\<rangle> P"
  shows "\<gamma> (t', a) \<inter> \<langle>G (t, a)\<rangle> P \<subseteq> \<langle>G (t', a)\<rangle> (dwp \<beta> P)"
  apply clarsimp
  apply(rule dwp_wp_transfer[where G="G (t, a)"])
  using assms apply blast
  using vmax_inclusion apply(fastforce simp add: \<beta>_def)
  by blast


lemma \<gamma>_act_dwp_wp_transfer2:
  assumes "P \<subseteq> \<langle>\<beta>\<rangle> P"
  shows "\<gamma> (t', a) \<inter> \<langle>G (t, a)\<rangle> P \<subseteq> \<langle>G (t', a)\<rangle> (\<gamma> (t', a) \<inter> dwp \<beta> P)"
  by (meson ViewTSys.\<gamma>_G_stable ViewTSys_axioms \<gamma>_\<beta>_stable \<gamma>_act_dwp_wp_transfer assms inf.cobounded1 interf_wp subset_trans wp_pred_int_pres)


lemma Vmax_inclusion:
  assumes "x \<in> \<Gamma> t' X"
      and "Y \<subseteq> X"
      and "(x, y') \<in> t' \<triangleright> Y"
    shows "\<exists>y z. (x, y) \<in> t \<triangleright> Y \<and> (y, z) \<in> \<beta> \<and> (y', z) \<in> \<beta>"
  using assms apply(unfold \<Gamma>_def Tr_def)
  apply clarsimp
  apply(subgoal_tac "xa \<in> X") defer  apply blast
  apply(subgoal_tac "\<exists>y. (x, y) \<in> G (t, xa) \<and>
                (\<exists>z. (y, z) \<in> \<beta> \<and> (y', z) \<in> \<beta>)")
   apply force
  apply(subgoal_tac "x \<in> \<gamma> (t', xa)") defer apply blast
  apply(drule vmax_inclusion[where y'=y' and t=t])
  apply(unfold \<beta>_def) by blast+

(*
lemma Vmax_\<beta>_confluence:
  assumes "x \<in> \<Gamma> t X"
      and "Y \<subseteq> X"
      and "(x, y) \<in> t \<triangleright> Y"
      and "(x, y') \<in> \<beta>"
    shows "\<exists> z. (y, z) \<in> \<beta> \<and> (y', z) \<in> t \<triangleright> Y"
  using assms apply(unfold \<Gamma>_def Tr_def)
  apply clarsimp
  apply(subgoal_tac "xa \<in> X") defer apply blast
  apply(subgoal_tac "\<exists>z. (y, z) \<in> \<beta> \<and> (y', z) \<in> G (t, xa)")
  apply force
  apply(subgoal_tac "x \<in> \<gamma> (t, xa)") defer apply blast
  apply(drule vmax_\<beta>_confluence[where y=y])
  apply(unfold \<beta>_def) by blast+
*)

lemma \<Gamma>_set_dwp_wp_transfer:
  assumes "P \<subseteq> \<langle>\<beta>\<rangle> P"
      and "Y \<subseteq> X"
    shows "\<Gamma> t' X \<inter> \<langle>t \<triangleright> Y\<rangle> P \<subseteq> \<langle>t' \<triangleright> Y\<rangle> (dwp \<beta> P)"
  apply clarsimp
  apply(rule dwp_wp_transfer[where G="t \<triangleright> Y"])
  using assms apply blast
  using assms Vmax_inclusion apply(fastforce simp add: \<beta>_def)
  by simp

lemma \<Gamma>_set_dwp_wp_transfer_full:
  assumes "P \<subseteq> \<langle>\<beta>\<rangle> P"
      and "Y \<subseteq> X"
      and "\<Gamma> t' X \<inter> dwp \<beta> P \<subseteq> Q"
    shows "\<Gamma> t' X \<inter> \<langle>t \<triangleright> Y\<rangle> P \<subseteq> \<langle>t' \<triangleright> Y\<rangle> Q"
  apply(rule dwp_wp_transfer_full[where B=\<beta>])
  using \<Gamma>_self_stable set_wp_left apply simp
  using assms(3) apply blast 
  using  \<Gamma>_set_dwp_wp_transfer[OF assms(1)]
  by (simp add: assms(2)) 

lemma emp_dwp_reduce:
  shows "dwp \<beta> {} \<subseteq> {}"
  by (simp add: dwp_emp)

(*
lemma \<Gamma>_set_dwp_reduce:
  assumes "P \<subseteq> \<langle>\<beta>\<rangle> P"
     and "Y \<subseteq> X"
  shows "\<Gamma> t' X \<inter> dwp \<beta> (\<langle>t \<triangleright> Y\<rangle> P) \<subseteq> \<langle>t' \<triangleright> Y\<rangle> (\<Gamma> t' X \<inter> dwp \<beta> P)"
  apply(rule dwp_wp_transfer2)
  using Vmax_\<beta>_confluence assms apply blast
  using \<beta>_trans apply blast
  apply (simp add: \<Gamma>_self_stable set_wp_left)
  using assms (1) apply (fastforce)
  using \<Gamma>_\<beta>_stable apply fastforce
  using Vmax_inclusion assms(2) by blast
*)

lemma int_dwp_reduce:
  shows "\<Gamma> t' X \<inter> dwp \<beta> (P \<inter> Q) \<subseteq> (\<Gamma> t' X \<inter> dwp \<beta> P) \<inter> (\<Gamma> t' X \<inter> dwp \<beta> Q)"
  apply(rule subset_trans[where B="\<Gamma> t' X \<inter> (dwp \<beta> P) \<inter> (dwp \<beta> Q)"])
   apply (metis (no_types, lifting) Int_mono dwp_int_pred_dist inf.cobounded1 inf_assoc inf_right_idem)
  by blast


lemma Int_dwp_reduce:
  shows "\<Gamma> t' X \<inter> dwp \<beta> (\<Inter> x \<in> A. P x) \<subseteq> (\<Inter> x \<in> A. (\<Gamma> t' X \<inter> dwp \<beta> (P x)))"
  apply(rule subset_trans[where B="\<Gamma> t' X \<inter> (\<Inter> x \<in> A. dwp \<beta> (P x))"])
  using dwp_Int_pred_dist apply fastforce
  by blast

lemma un_dwp_reduce:
  shows "\<Gamma> t' X \<inter> dwp \<beta> (P \<union> Q) \<subseteq> (\<Gamma> t' X \<inter> dwp \<beta> P) \<union> (\<Gamma> t' X \<inter> dwp \<beta> Q)"
  by (simp add: dwp_un_pred_dist inf_sup_distrib1)

lemma Un_dwp_reduce:
  shows "\<Gamma> t' X \<inter> dwp \<beta> (\<Union> x \<in> A. P x) \<subseteq> (\<Union> x \<in> A. (\<Gamma> t' X \<inter> dwp \<beta> (P x)))"
  using dwp_Un_pred_dist
  by (simp add: dwp_Un_pred_dist)


end


locale SharedVarsBase = ViewTSys +
  fixes var :: "'a \<Rightarrow> nat"
begin

definition "Var x = {a | a. var a = x}"

lemma var_of_Var:
   "Var y \<noteq> {} \<Longrightarrow> var`(Var y) = {y}"
  by(auto simp add: Var_def)

lemma [simp]:
  "Var y \<noteq> {} \<Longrightarrow> x \<in> var`(Var y) \<longleftrightarrow> x = y"
  by(simp add: var_of_Var)

end

locale SharedVars = SharedVarsBase +
  assumes any_max_to_synced:
    "var a = x \<Longrightarrow> \<gamma> (t, a) \<subseteq> \<langle>G (t, a)\<rangle> (\<Gamma> t (Var x))"
  assumes disinct_vars_\<gamma>_independent:
    "var a \<noteq> var b \<Longrightarrow> \<gamma> (t, a) \<subseteq> \<langle>interf (t', b)\<rangle> (\<gamma> (t, a))"
  assumes distinct_vars_act_independent:
    "var a \<noteq> var b \<Longrightarrow> interf (t', b) O G (t, a) \<subseteq> G (t, a) O interf (t', b)"
begin

lemma \<Gamma>_on_var_interf_stable:
  assumes "var b \<notin> var`X"
  shows "\<Gamma> t X \<subseteq> \<langle>interf (t', b)\<rangle> (\<Gamma> t X)"
  apply(rule \<gamma>_\<Gamma>_lift)
  apply(rule disinct_vars_\<gamma>_independent)
  using assms
  by (metis image_eqI)

lemma set_on_var_interf_stable:
  assumes "p \<subseteq> \<langle>interf (t', b)\<rangle> p"
    and "var b \<notin> var`X"
  shows "\<langle>t \<triangleright> X\<rangle> p \<subseteq> \<langle>interf (t', b)\<rangle> (\<langle>t \<triangleright> X\<rangle> p)"
  apply(rule wp_stable_set_cond)
  apply(rule wp_lipton)
   apply(rule distinct_vars_act_independent)
  using assms(2) apply (metis image_eqI)
  using assms(1) by simp

end

locale RWBase = SharedVars +
  fixes rval :: "'a \<Rightarrow> nat option"
    and wval :: "'a \<Rightarrow> nat option"
begin 


definition "Rds x v = {a | a. var a = x \<and> rval a = Some v \<and> wval a = None}"

lemma var_Rds:
  "Rds x v \<noteq> {} \<Longrightarrow> var`(Rds x v) = {x}"
  by(fastforce simp add: Rds_def)


lemma xvar_Rds [simp]:
  "Rds x v \<noteq> {} \<Longrightarrow> y \<in> var`(Rds x v) \<longleftrightarrow> y = x"
  by(fastforce simp add: Rds_def)



definition "not_value t x v \<equiv> \<langle>t \<triangleright> Rds x v\<rangle> {}"
definition "constrained_value t x V \<equiv> \<Inter> u \<in> {u . u \<notin> V}. not_value t x u"
definition "def_value t x v \<equiv> \<Inter> u \<in> {u . u \<noteq> v}. not_value t x u"
definition "synced t x v = def_value t x v \<inter> \<Gamma> t (Var x)"

lemma not_value_transfer:
  "\<Gamma> t' (Var x) \<inter> not_value t x v \<subseteq> not_value t' x v"
  apply(simp add: not_value_def)
  apply(subgoal_tac "Rds x v \<subseteq> Var x")
  using \<Gamma>_set_dwp_wp_transfer emp_dwp_reduce
  apply (metis empty_subsetI subset_empty)
  using Rds_def SharedVarsBase.Var_def SharedVarsBase_axioms by fastforce

lemma def_value_transfer:
  "\<Gamma> t' (Var x) \<inter> def_value t x v \<subseteq> def_value t' x v"
  apply(simp add: def_value_def)
  apply(rule subset_trans[where B="\<Inter> u \<in> {u. u \<noteq> v}. (\<Gamma> t' (Var x) \<inter> not_value t x u)"])
   apply blast
  using not_value_transfer by fastforce

lemma constrained_value_transfer:
  "\<Gamma> t' (Var x) \<inter> constrained_value t x V \<subseteq> constrained_value t' x V"
  apply(simp add: constrained_value_def)
  apply(rule subset_trans[where B="\<Inter> u \<in> {u. u \<notin> V}. (\<Gamma> t' (Var x) \<inter> not_value t x u)"])
   apply blast
  using not_value_transfer by fastforce

lemma synced_transfer:
  "\<Gamma> t' (Var x) \<inter> synced t x v \<subseteq> synced t' x v"
  apply(simp add: synced_def)
  using def_value_transfer by blast

end

locale RWVars = RWBase +
  assumes not_mod_\<gamma>_noninterfere:
    "wval r = None \<Longrightarrow> var r = var a \<Longrightarrow> \<gamma> (t, a) \<subseteq> \<langle>interf (t', r)\<rangle> (\<gamma> (t, a))"
  assumes not_mod_act_noninterfere:
    "wval r = None \<Longrightarrow>
     var a = var r \<Longrightarrow> interf (t', r) O G (t, a) \<subseteq> G (t, a) O interf (t', r)"

  assumes mod_distinct_vals_noninterfere:
    "wval r = None \<Longrightarrow>
     wval w = Some v \<Longrightarrow> rval r = Some v' \<Longrightarrow> v \<noteq> v' \<Longrightarrow>
     var r = var w \<Longrightarrow> interf (t', w) O G (t, r) \<subseteq> G (t, r) O interf (t', w)"

  assumes enabling_view_max:
    "var w = var r \<Longrightarrow> rval r \<noteq> None \<Longrightarrow> var a = var w \<Longrightarrow>
      wval w = rval r \<Longrightarrow>
         t \<noteq> t' \<Longrightarrow> (\<langle>G (t', r)\<rangle> {}) \<inter> \<gamma> (t, w) \<subseteq> \<langle>G (t, w)\<rangle> (\<langle>G (t', r)\<rangle> (\<gamma> (t', a)))"

  assumes read_enabled:
    "UNIV \<subseteq> (\<Union> v. Domain (t \<triangleright> Rds x v))"
  assumes mod_value_readable:
    "wval w = Some v \<Longrightarrow> UNIV \<subseteq> \<langle>G (t, w)\<rangle> (Domain (t \<triangleright> Rds (var w) v))"
  assumes weak_coherence:
    "\<Gamma> t (Var x) \<subseteq> (\<Union> v. def_value t x v)"
begin


lemma interf_read_is_\<beta>:
  assumes "wval r = None"
  shows "interf (t, r) \<subseteq> \<beta>"
  apply(subgoal_tac "interf (t, r) \<in> bndsim G {\<gamma> e | e. True} ")
  using \<beta>_def apply blast
  apply(simp add: \<beta>_def)
  apply(rule bndsim_intro)
  using not_mod_\<gamma>_noninterfere disinct_vars_\<gamma>_independent
   apply (smt CollectD assms) 
  apply(case_tac "var (snd e) = var r")
  using not_mod_act_noninterfere[OF assms(1)] apply fastforce
  using  distinct_vars_act_independent by fastforce


lemma read_is_\<beta>:
  assumes "wval r = None"
  shows "G (t, r) \<subseteq> \<beta>"
  using interference_valid interf_read_is_\<beta> assms
  by (metis (no_types, lifting) Sup_upper \<beta>_\<gamma>_bndsim \<beta>_def \<beta>_trans order_trans relcomp_mono)
  

lemma read_preserves_all:
  assumes "p \<subseteq> \<langle>\<beta>\<rangle> p"
   and "wval r = None"
  shows "p \<subseteq> \<langle>G (t, r)\<rangle> p"
  using read_is_\<beta>[OF assms(2)]
  by (metis assms(1) inf.absorb_iff1 inf.coboundedI2 wp_act_antimono2)

lemma read_set_preserves_all:
  assumes "p \<subseteq> \<langle>\<beta>\<rangle> p"
  shows "p \<subseteq> \<langle>t \<triangleright> Rds x v\<rangle> p"
  using read_preserves_all
  by (metis (mono_tags, lifting) CollectD Rds_def assms set_wp_left)

lemma read_set_enabled:
  "P \<subseteq> (\<Union> v. Domain (t \<triangleright> Rds x v))"
  apply(clarsimp)
  using read_enabled[where x=x and t=t] 
  by(fastforce simp add: Domain_Sup_distrib Tr_def Rds_def)

lemma mod_value_set_readable:
  assumes "wval w = Some v"
  shows "P \<subseteq> \<langle>G (t, w)\<rangle> (Domain (t \<triangleright> Rds (var w) v))"
  using mod_value_readable[OF assms(1)] 
  by(fastforce simp add: Domain_Sup_distrib Tr_def Rds_def)

lemma nonconflict_read_stable:
  assumes "P \<subseteq> \<langle>interf (t', a)\<rangle> P"
  assumes "var a \<noteq> x \<or> wval a \<noteq> Some v"
  shows "\<langle>t \<triangleright> Rds x v\<rangle> P \<subseteq> \<langle>interf (t', a)\<rangle> (\<langle>t \<triangleright> Rds x v\<rangle> P)"
  apply(case_tac "Rds x v = {}") apply simp
  apply(case_tac "var a = x")
  apply(rule wp_stable_set_cond)
  apply(case_tac "wval a \<noteq> Some v")
  apply(case_tac "wval a = None")
  apply(rule wp_lipton) 
  using distinct_vars_act_independent not_mod_act_noninterfere apply blast
  using assms(1) apply blast
  apply(rule wp_lipton) apply simp
  using mod_distinct_vals_noninterfere
  apply(simp add: Rds_def) apply metis
  using assms(1) apply blast
  using assms(2) apply blast
  apply(rule set_on_var_interf_stable[OF assms(1)])
  apply (simp add: Rds_def) by blast


lemma not_not_value_domain:
    "- not_value t x v = Domain (t \<triangleright> Rds x v)"
  by(fastforce simp add: not_value_def Rds_def wp_ext Tr_def)

lemma not_value_\<beta>_stable:
  shows "not_value t x v \<subseteq> \<langle>\<beta>\<rangle> (not_value t x v)"
  apply(unfold not_value_def)
  using set_\<beta>_stable by auto

lemma def_value_\<beta>_stable:
  shows "def_value t x v \<subseteq> \<langle>\<beta>\<rangle> (def_value t x v)"
  apply(unfold def_value_def)
  using not_value_\<beta>_stable
  by (simp add: Int_wp_pres2)

lemma constrained_value_\<beta>_stable:
  shows "constrained_value t x v \<subseteq> \<langle>\<beta>\<rangle> (constrained_value t x v)"
  apply(unfold constrained_value_def)
  using not_value_\<beta>_stable
  by (simp add: Int_wp_pres2)

lemma def_value_implies_not_value:
   assumes "v \<noteq> v'"
   shows "def_value t x v \<subseteq> not_value t x v'"
  apply(simp add: wp_ext def_value_def)
  apply(rule Inf_lower2[where u="not_value t x v'"])
  using assms by auto

lemma constrained_value_implies_not_value:
   assumes "v' \<notin> V"
   shows "constrained_value t x V \<subseteq> not_value t x v'"
  apply(simp add: wp_ext constrained_value_def)
  apply(rule Inf_lower2[where u="not_value t x v'"])
  using assms by auto

lemma def_value_unique:
  assumes "v \<noteq> v'"
  shows "def_value t x v \<inter> def_value t x v' = {}"
  apply(rule equals0I)
  apply(subgoal_tac "y \<in> not_value t x v")
   defer using def_value_implies_not_value assms
  apply (metis (no_types, lifting) IntE subsetD)
  apply(subgoal_tac "y \<in> not_value t x v'")
   defer using def_value_implies_not_value assms
   apply (metis (no_types, lifting) IntE subsetD)
  apply(subgoal_tac " \<exists> v''. y \<in> Domain (t \<triangleright> Rds x v'')")
  apply (smt ComplD IntD2 RWVars.def_value_implies_not_value RWVars_axioms not_not_value_domain subsetD)
  using read_set_enabled[where x=x and t=t and P=UNIV]
  using subsetD by fastforce

lemma not_value_stable:
  assumes "var a \<noteq> x \<or> wval a \<noteq> Some v"
  shows "not_value t x v \<subseteq> \<langle>interf (t', a)\<rangle> (not_value t x v)"
  apply(simp add: not_value_def)
  using assms(1) nonconflict_read_stable by auto

lemma def_value_stable:
  assumes "var a \<noteq> x \<or> wval a = Some v"
  shows "def_value t x v \<subseteq> \<langle>interf (t', a)\<rangle> (def_value t x v)"
  apply(unfold def_value_def)
  by (metis (mono_tags, lifting) CollectD Int_wp_pres2 assms not_value_stable option.inject)

lemma constrained_value_stable:
  assumes "var a \<noteq> x \<or> wval a \<in> Some `V"
  shows "constrained_value t x V \<subseteq> \<langle>interf (t', a)\<rangle> (constrained_value t x V)"
  apply(unfold constrained_value_def)
  apply(rule Int_wp_pres2 assms)
  using not_value_stable assms
  by (metis CollectD image_iff option.inject)

lemma synced_value_\<beta>_stable:
  "synced t x v \<subseteq> \<langle>\<beta>\<rangle> (synced t x v)"
  apply(simp add: synced_def)
  by (meson \<Gamma>_\<beta>_stable def_value_\<beta>_stable inf_le1 inf_le2 subset_trans wp_pred_int_pres)

lemma synced_value_stable:
  assumes "var a \<noteq> x \<or> (t' = t \<and> wval a = Some v)"
  shows "synced t x v \<subseteq> \<langle>interf (t', a)\<rangle> (synced t x v)"
  apply(unfold synced_def)
  apply(rule wp_ir_pr_conj_intro)
  apply(rule wp_ir_pr_conj_elim1)
  using assms def_value_stable apply blast
  apply(rule wp_ir_pr_conj_elim2)
  apply(case_tac "var a = x")
  using \<gamma>_G_stable \<gamma>_\<Gamma>_lift assms apply auto[1] 
  using \<Gamma>_on_var_interf_stable
  by (metis \<gamma>_\<Gamma>_lift empty_iff singletonD var_of_Var)


lemma def_value_implies_constrained_value:
  shows "def_value t x v \<subseteq> constrained_value t x {v}"
  by(auto simp add: def_value_def constrained_value_def)

lemma constrained_value_mono:
  assumes "V \<subseteq> V'"
  shows "constrained_value t x V \<subseteq> constrained_value t x V'"
  using assms by(auto simp add: constrained_value_def)

lemma def_value_self_write:
  assumes "wval w = Some v'"
  shows "\<Gamma> t (Var (var w)) \<subseteq> \<langle>G (t, w)\<rangle> (def_value t (var w) v')"
  apply rule
  apply(simp add: wp_ext)apply rule apply rule
  apply(subgoal_tac "s' \<in> \<Gamma> t (Var (var w))")
   defer using wp_pointwise_discharge[OF \<Gamma>_self_stable] apply blast
  apply(subgoal_tac "s' \<in> (\<Union> v. def_value t (var w) v)")
   defer using weak_coherence apply blast
  apply(subgoal_tac "s' \<in> Domain (t \<triangleright> Rds (var w) v')")
   defer using mod_value_set_readable apply (metis assms wp_pointwise_discharge)
  apply(unfold not_not_value_domain[symmetric])
  using def_value_implies_not_value 
  apply safe using subsetD by metis

lemma constrained_value_self_write:
  assumes "wval w \<in> Some`V"
  shows "\<Gamma> t (Var (var w)) \<subseteq> \<langle>G (t, w)\<rangle> (constrained_value t (var w) V)"
  apply(subgoal_tac "wval w = Some (the (wval w)) \<and> the(wval w) \<in> V")
  defer using assms apply fastforce
  apply(subgoal_tac "{the(wval w)} \<subseteq> V")
  defer apply fastforce
  using def_value_implies_constrained_value constrained_value_mono
  by (meson RWVars.def_value_self_write RWVars_axioms interpolation_right)

lemma synced_value_self_write:
  assumes "wval w = Some v'"
     and "var w = x"
  shows "synced t x v \<subseteq> \<langle>G (t, w)\<rangle> (synced t x v')"
  apply(simp add: synced_def)
  apply(rule wp_pred_int_pres)
  using def_value_self_write[OF assms(1)] assms(2) apply blast
  using \<Gamma>_self_stable by blast

lemma not_value_implies_wp:
  "not_value t x v \<subseteq> \<langle>t \<triangleright> Rds x v\<rangle> P"
  by (simp add: disabled_wp not_value_def)

lemma action_established_read:
  assumes "P \<subseteq> \<langle>A\<rangle> Q"
     and "Q \<subseteq> \<langle>\<beta>\<rangle> Q"
  shows "P \<subseteq> \<langle>A\<rangle> (\<langle>t \<triangleright> Rds x v\<rangle> Q)"
  apply(rule wp_pres_compos2[OF assms(1)])
  by(rule read_set_preserves_all[OF assms(2)])



lemma enabling_Vmax:
  assumes "var w = x"
      and "var r = x"
      and "rval r \<noteq> None"
      and "wval w = rval r" (* It would be good to relax this.*)
    shows " \<langle>G (t', r)\<rangle> {} \<inter> \<Gamma> t (Var x) \<subseteq> \<langle>G (t, w)\<rangle> (\<langle>G (t', r)\<rangle> (\<Gamma> t' (Var x)))"
  apply(cases "t = t'")
  using assms apply simp
   apply (metis \<Gamma>_self_stable inf.coboundedI2 interpolation_right)
  apply(cases "wval w = rval r")
   apply(simp add: \<Gamma>_def)
  apply(unfold wp_comp2 wp_comp[symmetric])
   apply(rule wp_pred_Int_pres_range)
   apply(rule subset_trans[where B="\<langle>G (t', r)\<rangle> {} \<inter> \<gamma> (t, w)"])
    apply(simp add: Var_def) using assms(1) apply blast
   apply(unfold wp_comp2[symmetric] wp_comp)
  using enabling_view_max assms apply simp 
   apply (metis (mono_tags, lifting) CollectD SharedVarsBase.Var_def SharedVarsBase_axioms)
  using assms by simp

lemma single_read_synced_single_var_transfer:
  assumes "var w = x"
      and "var r = x"
      and "wval w = Some v"
      and "rval r = Some v"
      and "wval r = None"
    shows "\<langle>G (t', r)\<rangle> {} \<inter> synced t x v' \<subseteq> \<langle>G (t, w)\<rangle> (\<langle>G (t', r)\<rangle> (synced t' x v))"
  apply(subgoal_tac "\<langle>G (t', r)\<rangle> {} \<inter> \<Gamma> t (Var x) \<subseteq> \<langle>G (t, w)\<rangle> (\<langle>G (t', r)\<rangle> (\<Gamma> t' (Var x)))")
  defer apply(rule enabling_Vmax[OF assms(1) assms(2)])
  apply (fastforce simp add: assms(4)) using assms apply force
  apply(subgoal_tac "synced t x v' \<subseteq> \<langle>G (t, w)\<rangle> (\<langle>G (t', r)\<rangle> (synced t x v))")
  defer apply (smt assms(1) assms(3) assms(5) interpolation_right read_preserves_all synced_value_\<beta>_stable synced_value_self_write)
  apply(rule subset_trans [where B= "\<langle>G (t, w)\<rangle> (\<langle>G (t', r)\<rangle> (\<Gamma> t' (Var x) \<inter> synced t x v))"])
  defer
  apply(rule wp_pred_mono) apply(rule wp_pred_mono[OF synced_transfer])
  apply(subst wp_comp2 wp_comp[symmetric])+
  apply(rule wp_ir_pr_conj_intro) apply(simp add: synced_def)
  apply fastforce 
  by fastforce


lemma act_Tr_set:
  assumes "a \<in> X"
  shows "G (t, a) \<subseteq> t \<triangleright> X"
  using assms by(auto simp add: Tr_def)

lemma act_Tr_set_wp:
  assumes "a \<in> X"
  shows "\<langle>t \<triangleright> X\<rangle> P \<subseteq> \<langle>G (t, a)\<rangle> P"
  using act_Tr_set
  by (simp add: assms wp_act_antimono2)

lemma enabling_Vmax_sync:
  assumes "var w = x"
      and "wval w = Some v"
    shows "\<langle>t' \<triangleright> Rds x v\<rangle> {} \<inter> synced t x v' \<subseteq> \<langle>G (t, w)\<rangle> (\<langle>t' \<triangleright> Rds x v\<rangle> (synced t' x v))"
  apply(subgoal_tac "\<And> r. var r = x \<and> rval r = Some v \<and> wval r = None \<Longrightarrow> \<langle>G (t', r)\<rangle> {} \<inter> synced t x v' \<subseteq> \<langle>G (t, w)\<rangle> (\<langle>G (t', r)\<rangle> (synced t' x v))")
  defer using single_read_synced_single_var_transfer
  using assms(1) assms(2) apply blast
   apply(case_tac "Rds x v = {}")
    apply fastforce
   apply(subgoal_tac "\<exists> r. r \<in> Rds x v")
  defer apply blast
   apply (erule exE) apply(auto simp add: wp_ext Tr_def Rds_def) (* TODO do something about this*)
proof -
fix r :: 'a and xa :: 'b and xaa :: 'a and s' :: 'b and s'a :: 'b and xb :: 'a
  assume a1: "\<And>ra. var ra = var r \<and> rval ra = Some v \<and> wval ra = None \<Longrightarrow> {s. \<forall>s'. (s, s') \<notin> G (t', ra)} \<inter> synced t (var r) v' \<subseteq> {s. \<forall>s'. (s, s') \<in> G (t, w) \<longrightarrow> (\<forall>s'a. (s', s'a) \<in> G (t', ra) \<longrightarrow> s'a \<in> synced t' (var r) v)}"
  assume a2: "xa \<in> synced t (var r) v'"
  assume a3: "\<forall>s' x. rval x = Some v \<longrightarrow> var x = var r \<longrightarrow> (\<exists>y. wval x = Some y) \<or> (xa, s') \<notin> G (t', x)"
  assume a4: "x = var r"
  assume a5: "(xa, s') \<in> G (t, w)"
  assume a6: "var xb = var r"
  assume a7: "rval xb = Some v"
  assume a8: "wval xb = None"
  assume a9: "(s', s'a) \<in> G (t', xb)"
  have f10: "var xb = x"
    using a6 a4 by meson
  have f11: "xa \<in> synced t x v'"
using a4 a2 by meson
  obtain nn :: "'a \<Rightarrow> nat" where
"\<forall>b. (xa, b) \<notin> G (t', xb) \<or> Some (nn xb) = wval xb"
    using f10 a7 a4 a3 by (metis (lifting))
then have f12: "\<forall>b. (xa, b) \<notin> G (t', xb)"
  using a8 by simp
  have "{b. \<forall>ba. (b, ba) \<notin> G (t', xb)} \<inter> synced t x v' \<subseteq> {b. \<forall>ba. (b, ba) \<in> G (t, w) \<longrightarrow> (\<forall>b. (ba, b) \<in> G (t', xb) \<longrightarrow> b \<in> synced t' (var r) v)}"
    using f10 a8 a7 a4 a1 by presburger
  then have "xa \<in> {b. \<forall>ba. (b, ba) \<in> G (t, w) \<longrightarrow> (\<forall>b. (ba, b) \<in> G (t', xb) \<longrightarrow> b \<in> synced t' (var r) v)}"
using f12 f11 by blast
  then show "s'a \<in> synced t' (var r) v"
using a9 a5 by blast
qed


lemma test_lemma:
  assumes "wval r = None"
      and "rval r = Some u"
      and "wval w = Some v"
      and "u \<noteq> v"
      and "P \<subseteq> \<langle>interf(t', w)\<rangle> P"
      and "P \<subseteq> \<langle>\<beta>\<rangle> P"
      and "var r = var w"
    shows "\<langle>G(t, r)\<rangle> P \<subseteq> \<langle>G(t', w)\<rangle>(\<langle>G(t, r)\<rangle> P)"
  apply(rule interf_wp)
   apply (simp add: G_\<beta>_stable assms(6))
  by (metis assms(1) assms(2) assms(3) assms(4) assms(5) assms(7) mod_distinct_vals_noninterfere wp_lipton)

end

context RWVars begin

lemma initial_not_value:
  assumes "v \<noteq> v'"
  shows "synced t f v \<subseteq> \<langle>t \<triangleright> Rds f v'\<rangle> {}"
  using assms def_value_implies_not_value not_value_def synced_def by fastforce

end


locale RMWAtomicity = RWVars +
  fixes up_rd up_wr
  assumes rmw_is_rw:
    "wval u \<noteq> None \<Longrightarrow> rval u \<noteq> None \<Longrightarrow> G (t', u) \<subseteq> (G (t, up_rd u)) O (G (t, up_wr u)) O \<beta>"
  assumes update_read:
    "rval u \<noteq> None \<Longrightarrow> wval u \<noteq> None \<Longrightarrow> rval(up_rd u) = rval u \<and> wval(up_rd u) = None"
  assumes update_write:
    "rval u \<noteq> None \<Longrightarrow> wval u \<noteq> None \<Longrightarrow> wval(up_wr u) = wval u \<and> rval(up_wr u) = None"
  assumes update_vars:
    "rval u \<noteq> None \<Longrightarrow> wval u \<noteq> None \<Longrightarrow> var(up_wr u) = var(up_rd u)"
  assumes covering:
    "var w = var u \<Longrightarrow> wval w \<noteq> None \<Longrightarrow> wval u \<noteq> None \<Longrightarrow> rval u \<noteq> None \<Longrightarrow>
          t' \<noteq> t \<Longrightarrow> \<gamma> (t, w) \<subseteq> \<langle>G (t', u)\<rangle> (\<gamma> (t, w))"
begin


lemma RMW_wp_dcomposition:
  assumes "Q \<subseteq> \<langle>\<beta>\<rangle> Q"
   and "wval u \<noteq> None"
   and "rval u \<noteq> None"
   and "P \<subseteq> \<langle>G (t, up_rd u)\<rangle> (\<langle>G (t, up_wr u)\<rangle> Q)"
 shows "P \<subseteq> \<langle>G(t, u)\<rangle> Q"
  apply(rule subset_trans[where B="\<langle>G(t, up_rd u)\<rangle> (\<langle>G (t, up_wr u)\<rangle> (\<langle>\<beta>\<rangle> Q))"])
  apply (metis (no_types, lifting) assms(1) assms(4) inf.absorb_iff1 inf.coboundedI2 wp_pred_mono)
  apply(unfold wp_comp2 wp_comp[symmetric])
  apply(rule wp_act_antimono2) using assms rmw_is_rw 
  by simp

lemma RMW_wp_dcomposition2:
  assumes "Q \<subseteq> \<langle>\<beta>\<rangle> Q"
  assumes "P \<subseteq> \<langle>\<beta>\<rangle> P"
   and "wval u \<noteq> None"
   and "rval u \<noteq> None"
   and "P \<subseteq> \<langle>G (t, up_wr u)\<rangle> Q"
 shows "P \<subseteq> \<langle>G(t, u)\<rangle> Q"
  apply(rule RMW_wp_dcomposition[OF assms(1) assms(3) assms(4)])
  by (smt G_\<beta>_stable RMWAtomicity.update_read RMWAtomicity_axioms assms(1) assms(3) assms(4) assms(5) inf.absorb_iff1 inf.coboundedI2 read_preserves_all)


end


locale TSOFences = RWVars +
  fixes fence
  assumes fence_wval:
    "wval fence = None"
  assumes fencing:
    "t \<noteq> t' \<Longrightarrow> \<gamma> (t, c) \<subseteq> \<langle>G (t, fence)\<rangle> (\<gamma> (t', c))"
begin

lemma set_fencing:
  shows "\<Gamma> t X \<subseteq> \<langle>G (t, fence)\<rangle> (\<Gamma> t' X)"
  apply(simp add:_\<Gamma>_def) using fencing
  by (smt Inf_tool \<Gamma>_self_stable \<gamma>_is_\<Gamma> image_iff wp_pred_Int_pres)

lemma sync_fencing:
  shows "synced t x v \<subseteq> \<langle>G (t, fence)\<rangle> (synced t' x v)"
  using set_fencing[where X="Var x"]
  apply(subgoal_tac "synced t x v \<subseteq> \<langle>G (t, fence)\<rangle> (synced t x v)")
   defer using fence_wval apply (simp add: read_preserves_all synced_value_\<beta>_stable) (* should complete *)
  apply(subgoal_tac "synced t x v \<subseteq> \<langle>G (t, fence)\<rangle> (\<Gamma> t' (Var x))")
   defer apply(simp add: synced_def) apply (fastforce simp add: inf.coboundedI2)
  by (smt Int_subset_iff RWBase.def_value_transfer RWBase_axioms interpolation_right synced_def wp_pred_int_distrib)

end


locale SyncVars = RWVars +
  fixes acquiring releasing
  assumes sync:
    "var w = var r \<Longrightarrow> var w \<noteq> var c \<Longrightarrow> rval r \<noteq> None \<Longrightarrow>
      wval w = rval r \<Longrightarrow> (w, r) \<in> releasing \<times> acquiring  \<Longrightarrow>
          t \<noteq> t' \<Longrightarrow>
              (\<langle>G (t', r)\<rangle> (\<gamma> (t', c))) \<inter> \<gamma> (t, c) \<subseteq> \<langle>G (t, w)\<rangle> (\<langle>G (t', r)\<rangle> (\<gamma> (t', c)))"
begin


lemma \<gamma>_sync_other_var:
  assumes "var w = var r \<longrightarrow> wval w = Some (the (rval r)) \<longrightarrow> (w, r) \<in> releasing \<times> acquiring"
     and "rval r \<noteq> None"
     and "wval r = None"
     and "var w \<noteq> var c"
  shows "(\<langle>G (t', r)\<rangle> (\<gamma>  (t', c))) \<inter> \<gamma> (t, c) \<subseteq> \<langle>G (t, w)\<rangle> (\<langle>G (t', r)\<rangle> (\<gamma> (t', c)))"
  apply(case_tac "t=t'") apply simp
  apply (metis \<Gamma>_self_stable \<gamma>_is_\<Gamma> inf.coboundedI2 wp_comp wp_comp2 wp_pres_compose)
  apply(case_tac "var w \<noteq> var r")
  apply(case_tac "var w \<noteq> var c")
  using \<Gamma>_self_stable \<gamma>_is_\<Gamma> le_infI1 set_wp_left
  apply (metis (mono_tags, lifting) G_\<beta>_stable SharedVars.disinct_vars_\<gamma>_independent SharedVars_axioms \<gamma>_\<beta>_stable distinct_vars_act_independent interf_wp wp_lipton)
  using assms(4) apply blast
  apply simp
  apply(case_tac "wval w = None")
   apply (simp add: G_\<beta>_stable \<gamma>_\<beta>_stable le_infI1 read_preserves_all)
  apply(case_tac "rval r = Some (the (wval w))")
  using assms(1) assms(4) sync apply fastforce
  apply(rule subset_trans[where B="\<langle>G (t', r)\<rangle> (\<gamma> (t', c))"])
  apply blast
  apply(rule interf_wp) using G_\<beta>_stable \<gamma>_\<beta>_stable apply blast
  apply(subgoal_tac "\<gamma> (t', c) \<subseteq> \<langle>interf (t, w)\<rangle> (\<gamma> (t', c))")
   defer apply (metis SharedVars.disinct_vars_\<gamma>_independent SharedVars_axioms assms(4))
  apply(rule wp_lipton[OF mod_distinct_vals_noninterfere[where v="the(wval w)" and v'="the(rval r)"]])
  using assms(3) apply blast apply fastforce
  using assms(2) apply fastforce
  using assms(2) apply auto[1] apply fastforce
  by blast

lemma \<Gamma>_set_sync_other_var:
  assumes "var w = var r \<longrightarrow> wval w = Some (the (rval r)) \<longrightarrow> (w, r) \<in> releasing \<times> acquiring"
      and "rval r \<noteq> None"
      and "wval r = None"
      and "var w \<notin> var`X"
      and "X \<subseteq> Y"
    shows "(\<langle>G (t', r)\<rangle> (\<Gamma> t' X)) \<inter> \<Gamma> t Y \<subseteq> \<langle>G (t, w)\<rangle> (\<langle>G (t', r)\<rangle> (\<Gamma> t' X))"
  apply(simp add: \<Gamma>_def)
  using \<gamma>_sync_other_var[OF assms(1) assms(2) assms(3)]
  using assms(4,5) (* Need better proof. *)
  apply(auto simp add: wp_ext) by blast

lemma wp_pres_absorb:
  assumes "q \<subseteq> \<langle>a\<rangle> q"
  shows "(\<langle>a\<rangle> p) \<inter> q \<subseteq> \<langle>a\<rangle> (p \<inter> q)"
  by (metis Int_subset_iff assms inf.absorb_iff2 inf_le2 wp_pred_int_pres)

lemma \<Gamma>_set_synced:
  assumes "var w = var r \<longrightarrow> wval w = Some (the (rval r)) \<longrightarrow> (w, r) \<in> releasing \<times> acquiring"
      and "rval r \<noteq> None"
      and "wval r = None"
      and "y \<noteq> var w"
    shows "(\<langle>G (t', r)\<rangle> (synced t' y v)) \<inter> synced t y v \<subseteq> \<langle>G (t, w)\<rangle> (\<langle>G (t', r)\<rangle> (synced t' y v))"
  apply(subgoal_tac "synced t y v \<subseteq> \<langle>G (t, w)\<rangle> (synced t y v)")
    defer using assms(4) interf_wp synced_value_stable synced_value_\<beta>_stable apply metis
  apply(subgoal_tac "(\<langle>G (t', r)\<rangle> (synced t' y v)) \<inter> synced t y v \<subseteq> \<langle>G (t, w)\<rangle> (\<langle>G (t', r)\<rangle> (\<Gamma> t' (Var y)))")
  apply(rule subset_trans[where B="\<langle>G (t, w)\<rangle> (\<langle>G (t', r)\<rangle> (\<Gamma> t' (Var y)) \<inter> synced t y v)"])
  apply (smt Int_subset_iff inf.absorb_iff2 inf_le2 wp_pred_int_pres)
   apply(rule wp_pred_mono) using wp_pres_absorb
  apply (metis (no_types, lifting) assms(3) interpolation_right read_preserves_all synced_transfer synced_value_\<beta>_stable)
  apply(rule subset_trans[where B="(\<langle>G (t', r)\<rangle> (\<Gamma> t' (Var y))) \<inter> \<Gamma> t (Var y)"])
   apply (metis Int_mono inf_le2 synced_def wp_pred_mono)
  by (smt \<Gamma>_set_sync_other_var assms(1) assms(2) assms(3) assms(4) empty_iff image_iff set_eq_subset singletonD var_of_Var)

end


locale RWVarsActions = RWBase +
  fixes R W U F
  assumes
        R_var [simp]: "var (R b x v) = x"
    and W_var [simp]: "var (W (b :: bool) x v) = x"
    and U_var [simp]: "var (U (br :: bool) (bw :: bool) x v v') = x"
    and R_rval [simp]: "rval (R b x v) = Some v"
    and R_wval [simp]: "wval (R b x v) = None"
    and W_rval [simp]: "rval (W b x v) = None"
    and W_wval [simp]: "wval (W b x v) = Some v"
    and U_rval [simp]: "rval (U br bw x v v') = Some v"
    and U_wval [simp]: "wval (U br bw x v v') = Some v'"
    and F_rval [simp]: "rval F = None"
    and F_wval [simp]: "wval F = None"
begin

lemma R_Rds: "R b x v \<in> Rds x v" by(simp add: Rds_def)

definition "acquiring = {a | a x v. a = R True x v \<or> (\<exists> bw v'. a = U True bw x v v')}"
definition "releasing = {a | a x v. a = W True x v \<or> (\<exists> br v'. a = U br True x v v')}"

end


locale MPCase =
  acts: RWVarsActions + SyncVars G Init \<gamma> interf var rval wval acts.acquiring acts.releasing


locale SC = RWVars +
  assumes SC_vmax: "UNIV \<subseteq> \<gamma> e"
begin

lemma SC_sync:
 "var w = var r \<Longrightarrow>
       var w \<noteq> var c \<Longrightarrow>
       rval r \<noteq> None \<Longrightarrow>
       wval w = Some (the (rval r)) \<Longrightarrow>
       (w, r) \<in> UNIV \<times> UNIV \<Longrightarrow>
       t \<noteq> t' \<Longrightarrow> \<langle>G (t', r)\<rangle> (\<gamma> (t', c)) \<inter> \<gamma> (t, c) \<subseteq> \<langle>G (t, w)\<rangle> (\<langle>G (t', r)\<rangle> (\<gamma> (t', c)))"
  apply(rule subset_trans[where B="\<langle>G (t, w)\<rangle> (\<langle>G (t', r)\<rangle> UNIV)"])
   apply fastforce
  using SC_vmax wp_pred_mono
  by metis

end

sublocale SC \<subseteq> SyncVars G Init \<gamma> interf var rval wval UNIV UNIV
  apply(unfold_locales)
  apply simp 
  by (simp add: SC_sync)

end

