theory WP_SP
  imports Diagonals
begin

lemma fun_ext: "(\<And> x. f x = g x) \<Longrightarrow> f = g"
  by fastforce

(* The weakest precondition such that A achieves P *)
definition wp :: "'a rel \<Rightarrow> 'a set \<Rightarrow> 'a set" ( "(\<langle>_\<rangle>)" [10])
  where
  "wp A P \<equiv> \<Union> {S . A``S \<subseteq> P}" 

lemma wp_ext: "\<langle>A\<rangle> P = {s . (\<forall> s'. (s, s') \<in> A \<longrightarrow> s' \<in> P)}"
  apply(simp add: wp_def) apply rule
   apply blast
  apply (rule subsetI) apply safe
  by (meson CollectI Image_singleton_iff Union_iff singletonI subsetI)


lemma wp_act_antimono: "A \<subseteq> A' \<Longrightarrow> \<langle>A'\<rangle> \<le> \<langle>A\<rangle>"
  apply (simp add: le_fun_def wp_ext) by blast

lemma wp_act_antimono2: "A \<subseteq> A' \<Longrightarrow> \<langle>A'\<rangle> P \<subseteq>  \<langle>A\<rangle> P"
  by (simp add: wp_act_antimono le_funD)


lemma wp_pred_mono: "P \<subseteq> P' \<Longrightarrow> \<langle>A\<rangle> P \<subseteq> \<langle>A\<rangle> P'"
  apply (simp add: wp_ext) by blast

lemma wp_order:
  assumes "\<langle>A\<rangle> \<le> \<langle>B\<rangle>"
  shows "\<langle>A\<rangle> P \<subseteq> \<langle>B\<rangle> P"
  by(rule le_funD[OF assms])

lemma wp_comp [simp]:
  shows "\<langle>A O B\<rangle> = \<langle>A\<rangle> \<circ> \<langle>B\<rangle>"
  apply(rule fun_ext)
  apply(simp add: wp_ext) by blast


lemma wp_comp2: (* Do not make this a simp. *)
  shows "\<langle>A\<rangle> (\<langle>B\<rangle> P) = (\<langle>A\<rangle> \<circ> \<langle>B\<rangle>) P"
  by simp

lemma wp_discharge1:
  shows "A``(\<langle>A\<rangle> P) \<subseteq> P"
  using wp_def by fastforce

lemma wp_discharge:
  assumes "R \<subseteq> A"
  shows "R``(\<langle>A\<rangle> P) \<subseteq> P"
  using wp_discharge1 
  by (metis Image_mono assms order_refl order_trans)

lemma wp_sp_galois:
  shows "(P \<subseteq> \<langle>R\<rangle> Q) = (R``P \<subseteq> Q)"
  apply (simp add: Sup_upper wp_def) by auto

lemma wp_Id [simp]:
  "\<langle>Id\<rangle> P = P"
  by(auto simp add: wp_ext)

lemma wp_lipton_degenerate:
  assumes "R O A \<subseteq> A"
  shows "\<langle>A\<rangle> \<le> \<langle>R\<rangle> \<circ> \<langle>A\<rangle>"
  using assms(1) wp_act_antimono wp_comp by metis

lemma wp_star_induct:
  assumes "X \<subseteq> Y"
   and "X \<subseteq> \<langle>R\<rangle> X"
  shows "X \<subseteq> \<langle>R\<^sup>*\<rangle> Y"
  using assms apply(simp add: wp_ext) apply clarsimp
  by (metis (no_types, lifting) ImageI Image_closed_trancl assms(2) subsetD wp_sp_galois)

lemma wp_pres_star:
  "P \<subseteq> \<langle>A\<rangle> P \<Longrightarrow> P \<subseteq> \<langle>A\<^sup>*\<rangle> P"
  by (meson order_refl wp_sp_galois wp_star_induct)

lemma wp_interference_expand:
  assumes "P \<subseteq> \<langle>R\<rangle> P"
  shows "\<langle>A\<rangle> P \<subseteq> \<langle>A\<rangle> (\<langle>R\<^sup>*\<rangle> P)"
  using assms
  by (simp add: wp_pred_mono wp_pres_star)

lemma wp_star_pres:
  shows "\<langle>R\<^sup>*\<rangle> \<le> \<langle>R\<rangle> \<circ> \<langle>R\<^sup>*\<rangle>"
  apply(rule wp_lipton_degenerate) by fastforce

lemma wp_semi_comm:
  "R O A \<subseteq> A' O R' \<Longrightarrow> \<langle>A'\<rangle> \<circ> \<langle>R'\<rangle> \<le> \<langle>R\<rangle> \<circ> \<langle>A\<rangle>"
  using wp_act_antimono wp_comp by metis
  
lemma wp_lipton_gen:
  assumes "R O A \<subseteq> A' O R'"
      and "P \<subseteq> \<langle>R'\<rangle> P'"
    shows "\<langle>A'\<rangle> P \<subseteq> \<langle>R\<rangle> (\<langle>A\<rangle> P')"
  apply(rule subset_trans[where B="\<langle>A'\<rangle> (\<langle>R'\<rangle> P')"])
   apply(rule wp_pred_mono[OF assms(2)])
  apply(unfold wp_comp2)
  by(rule le_funD[OF wp_semi_comm[OF assms(1)]])

lemma wp_lipton:
  assumes "R O A \<subseteq> A O R"
      and "P \<subseteq> \<langle>R\<rangle> P"
    shows "\<langle>A\<rangle> P \<subseteq> \<langle>R\<rangle> (\<langle>A\<rangle> P)"
  by(rule wp_lipton_gen[OF assms])

lemma wp_pres_compose:
  assumes "P \<subseteq> wp r Q"
      and "Q \<subseteq> wp s R"
    shows "P \<subseteq> wp (r O s) R"
  using assms(1) assms(2) dual_order.trans wp_comp wp_pred_mono by fastforce

lemma wp_pres_compos2:
  assumes "P \<subseteq> wp r Q"
      and "Q \<subseteq> wp s R"
    shows "P \<subseteq> \<langle>r\<rangle> (\<langle>s\<rangle> R)"
  using assms(1) assms(2) dual_order.trans wp_comp wp_pred_mono by fastforce

lemma wp_Sup:
  assumes "\<And> v. P \<subseteq> wp (r v) Q"
  shows "P \<subseteq> wp (\<Union> v. r v) Q"
  using assms by(auto simp add: wp_ext)

lemma wp_Sup2:
  assumes "\<And> r. r \<in> rs \<Longrightarrow> P \<subseteq> wp (r) Q"
  shows "P \<subseteq> wp (\<Union> rs) Q"
  using assms by(auto simp add: wp_ext)

lemma sp_sup_int:
  "wp (A \<union> B) P = wp A P \<inter> wp B P" by(auto simp add: wp_ext)

lemma wp_Sup_Int:
  "wp (\<Union>A) P = (\<Inter>a\<in>A. (wp a P))"
  by (auto simp add: wp_ext)

lemma wp_kripke:
  "wp A (-X \<union> Y) \<subseteq> -(wp A X) \<union> wp A Y"
  by(auto simp add: wp_ext)

lemma wp_kripke1:
  "wp A (X \<union> Y) \<subseteq> -(wp A (-X)) \<union> wp A Y"
  using wp_kripke by force

lemma set_dichotomy:
  assumes " X \<inter> Y \<subseteq> Z"
      and "-X \<inter> Y \<subseteq> Z"
  shows "Y \<subseteq> Z"
  using assms(1) assms(2) by blast

lemma [simp]: "\<langle>A\<rangle> {} \<inter> \<langle>A\<rangle> P = \<langle>A\<rangle> {}"
  by (simp add: Int_absorb2 wp_pred_mono)

lemma disabled_wp: "\<langle>A\<rangle> {} \<subseteq> \<langle>A\<rangle> P"
  by (simp add: wp_pred_mono)

lemma disabled_wp2: "B \<subseteq> A \<Longrightarrow> \<langle>A\<rangle> {} \<subseteq> \<langle>B\<rangle> P"
  by (auto simp add: wp_ext)

lemma wp_pred_int_distrib:
  "wp R (X \<inter> Y) = wp R X \<inter> wp R Y"
  by(auto simp add: wp_ext)

lemma wp_act_int_semi_distrib:
  "wp R X \<inter> wp S X \<subseteq> wp (R \<inter> S) X"
  by(auto simp add: wp_ext)

lemma wp_pred_sup_semi_distrib:
  "wp R X \<union> wp R Y \<subseteq> wp R (X \<union> Y)"
  by(auto simp add: wp_ext)

lemma wp_act_sup_semi_distrib:
  "wp (R \<union> S) X \<subseteq> wp R X \<union> wp S X"
  by(auto simp add: wp_ext)

lemma wp_pred_Int_pres:
  assumes "\<And> x. x \<in> X \<Longrightarrow> P \<subseteq> wp R x"
  shows "P \<subseteq> wp R (\<Inter> X)"
  using assms by(auto simp add: wp_ext)

lemma wp_pred_Int_pres_range:
  assumes "\<And> x. x \<in> X \<Longrightarrow> P \<subseteq> wp R (Q (w, x))"
  shows "P \<subseteq> wp R (\<Inter> a \<in> X. Q (w, a))"
  using assms by(auto simp add: wp_ext)

lemma wp_pred_int_pres:
  assumes "P \<subseteq> wp R X"
      and "P \<subseteq> wp R Y"
  shows "P \<subseteq> wp R (X \<inter> Y)"
  using assms by(auto simp add: wp_ext)

lemma wp_pred_Union_pres:
  assumes "\<And> x. x \<in> X \<Longrightarrow> x \<subseteq> wp R P"
  shows "\<Union> X \<subseteq> wp R P"
  using assms by(auto simp add: wp_ext)

lemma wp_pred_union_pres:
  assumes "X \<subseteq> wp R P"
     and "Y \<subseteq> wp R P"
  shows "X \<union> Y \<subseteq> wp R P"
  using assms by(auto simp add: wp_ext)

lemma interpolation_left:
  assumes "P' \<subseteq> wp R Q"
     and "P \<subseteq> P'"
  shows "P \<subseteq> wp R Q"
  using assms(1) assms(2) by auto

lemma interpolation_right:
  assumes "P \<subseteq> wp R Q'"
     and "Q' \<subseteq> Q"
  shows "P \<subseteq> wp R Q"
  using assms(1) assms(2) wp_pred_mono by force

lemma Range_Act_wp:
  shows "wp A (Range A) = UNIV"
  by(auto simp add: wp_ext)

lemma Range_pres:
  assumes "R O A \<subseteq> A O R"
  shows "Range R \<subseteq> wp A (Range R)"
  using assms wp_ext by fastforce

lemma Int_wp_pres:
  assumes "\<And> a. p a \<subseteq> wp A (p a)"
  shows "(\<Inter>a. p a) \<subseteq> wp A (\<Inter>a. p a)"
  by (smt Inter_lower assms dual_order.trans imageE wp_pred_Int_pres)

lemma Int_wp_pres2:
  assumes "\<And> a.  a \<in> X \<Longrightarrow> p a \<subseteq> wp A (p a)"
  shows "(\<Inter>a \<in> X. p a) \<subseteq> wp A (\<Inter>a \<in> X. p a)"
  by (smt Inter_lower assms imageE interpolation_left wp_pred_Int_pres)

lemma Int_wp_pres3:
  assumes "\<And> q. q \<in> Q \<Longrightarrow> p \<subseteq> wp A q"
   and "p \<in> P"
  shows "(\<Inter>P) \<subseteq> wp A (\<Inter>Q)"
  by (meson Inter_lower assms(1) assms(2) interpolation_left wp_pred_Int_pres)

lemma pres_intro:
  assumes "\<And> x y. x \<in> P \<Longrightarrow> (x, y) \<in> R \<Longrightarrow> y \<in> Q"
  shows "P \<subseteq> wp R Q"
  by (simp add: assms subsetI wp_ext)

definition "csubs r r' = {x . r``{x} \<subseteq> r'``{x}}"

lemma csubs_refl:
  "UNIV \<subseteq> csubs r r"
  by(auto simp add: csubs_def)

lemma csubs_refl2:
  "csubs r r = UNIV"
  by(auto simp add: csubs_def)

lemma csubs_refl3:
  "r \<subseteq> r' \<Longrightarrow> csubs r r' = UNIV"
  by(auto simp add: csubs_def)

lemma csubs_trans:
  "csubs r s \<inter> csubs s t \<subseteq> csubs r t"
  by(auto simp add: csubs_def)

lemma csubs_star:
  "csubs r (s O t\<^sup>*) \<subseteq> csubs (r O t\<^sup>*) (s O t\<^sup>*)"
  by(fastforce simp add: csubs_def)

lemma csubs_fst_antimono:
  assumes "A \<subseteq> A'"
  shows "csubs A' B \<subseteq> csubs A B"
  using assms by(auto simp add: csubs_def)

lemma csubs_snd_mono:
  assumes "B \<subseteq> B'"
  shows "csubs A B \<subseteq> csubs A B'"
  using assms by(auto simp add: csubs_def)

lemma csubs_transfer:
  shows "\<langle>B\<rangle> P \<inter> csubs A B \<subseteq> \<langle>A\<rangle> P"
  by(auto simp add: wp_ext csubs_def)

lemma csubs_wp_pres:
  assumes "\<And> x y z. x \<in> csubs A B \<and> (x, y) \<in> C \<and> (y, z) \<in> A \<Longrightarrow> (y, z) \<in> B"
  shows "csubs A B \<subseteq> wp C (csubs A B)"
  using assms by(auto simp add: wp_ext csubs_def)

lemma weakest_action_correct:
  "P \<subseteq> \<langle>\<Union> {A . P \<subseteq> \<langle>A\<rangle> Q}\<rangle>  Q"
  using wp_Sup_Int by fastforce



(*

Natural deduction style rules for wp. Obviously,
several of these are essentially immediate consequences
of earlier lemmas. I have them here for sake of having
a complete set of rules.

*)

lemma wp_ir_pr_disj_elim:
  assumes "p \<subseteq> \<langle>a\<rangle> q"
      and "p' \<subseteq> \<langle>a\<rangle> q"
    shows" p \<union> p' \<subseteq> \<langle>a\<rangle> q"
  using assms by(auto simp add: wp_ext)

lemma wp_ir_pr_disj_intro1:
  assumes "p \<subseteq> \<langle>a\<rangle> q"
    shows" p \<subseteq> \<langle>a\<rangle> (q \<union> q')"
  using assms by(auto simp add: wp_ext)

lemma wp_ir_pr_disj_intro2:
  assumes "p \<subseteq> \<langle>a\<rangle> q'"
    shows" p \<subseteq> \<langle>a\<rangle> (q \<union> q')"
  using assms by(auto simp add: wp_ext)

lemma wp_ir_pr_conj_elim1:
  assumes "p \<subseteq> \<langle>a\<rangle> q"
    shows" p \<inter> p' \<subseteq> \<langle>a\<rangle> q"
  using assms by(auto simp add: wp_ext)

lemma wp_ir_pr_conj_elim2:
  assumes "p' \<subseteq> \<langle>a\<rangle> q"
    shows" p \<inter> p' \<subseteq> \<langle>a\<rangle> q"
  using assms by(auto simp add: wp_ext)

lemma wp_ir_pr_conj_intro:
  assumes "p \<subseteq> \<langle>a\<rangle> q"
      and "p \<subseteq> \<langle>a\<rangle> q'"
    shows" p \<subseteq> \<langle>a\<rangle> (q \<inter> q')"
  using assms by(auto simp add: wp_ext)

lemma wp_ir_pr_Int_intro:
  assumes "\<And> q. q \<in> Q \<Longrightarrow> p \<subseteq> \<langle>a\<rangle> q"
  shows "p \<subseteq> \<langle>a\<rangle> (\<Inter> Q)"
  using assms by(auto simp add: wp_ext)

lemma wp_ir_pr_Int_elim:
  assumes "p \<in> P"
      and "p \<subseteq> \<langle>a\<rangle> q"
    shows "(\<Inter> P) \<subseteq> \<langle>a\<rangle> q"
  using assms by(auto simp add: wp_ext)

lemma wp_ir_pr_Un_intro:
  assumes "q \<in> Q"
     and "p \<subseteq> \<langle>a\<rangle> q"
  shows "p \<subseteq> \<langle>a\<rangle> (\<Union> Q)"
  using assms by(auto simp add: wp_ext)

lemma wp_ir_pr_Un_elim:
  assumes "\<And> p. p \<in> P \<Longrightarrow> p \<subseteq> \<langle>a\<rangle> q"
  shows "(\<Union> P) \<subseteq> \<langle>a\<rangle> q"
  using assms by(auto simp add: wp_ext)

lemma wp_ir_pr_bot:
  shows "{} \<subseteq> \<langle>a\<rangle> q"
  by(auto simp add: wp_ext)

lemma wp_ir_pr_top:
  shows "p \<subseteq> \<langle>a\<rangle> UNIV"
  by(auto simp add: wp_ext)

lemma wp_ir_pr_ncd:
  assumes "r \<inter> p \<subseteq> \<langle>a\<rangle> q"
     and "-r \<inter> p \<subseteq> \<langle>a\<rangle> q"
   shows "p \<subseteq> \<langle>a\<rangle> q"
  using assms(1) assms(2) by blast

lemmas wp_ir_pr = wp_ir_pr_disj_elim wp_ir_pr_disj_intro1 wp_ir_pr_disj_intro2 wp_ir_pr_conj_elim1
wp_ir_pr_conj_elim2 wp_ir_pr_conj_intro wp_ir_pr_Int_intro wp_ir_pr_Int_elim
wp_ir_pr_Un_intro wp_ir_pr_Un_elim wp_ir_pr_bot wp_ir_pr_top wp_ir_pr_ncd

lemma wp_ir_act_bot:
  shows "p \<subseteq> q \<longleftrightarrow> p \<subseteq> \<langle>Id\<rangle> q"
  by(auto simp add: wp_ext)

lemma wp_ir_act_pred_mono:
  assumes "p \<subseteq> q"
  shows "\<langle>a\<rangle> p \<subseteq> \<langle>a\<rangle> q"
  using assms by(auto simp add: wp_ext)

lemma wp_ir_act_act_mono:
  assumes "a \<subseteq> b"
  shows "\<langle>b\<rangle> p \<subseteq> \<langle>a\<rangle> p"
  using assms by(auto simp add: wp_ext)

(* What follows are three ways to prove action
  intersection. These are sufficient but not necessary,
  even in combination. *)

lemma wp_ir_act_int1:
  assumes "p \<subseteq> \<langle>a\<rangle> q"
    shows "p \<subseteq> \<langle>a \<inter> b\<rangle> q"
  using assms by(auto simp add: wp_ext)

lemma wp_ir_act_int2:
  assumes "p \<subseteq> \<langle>b\<rangle> q"
    shows "p \<subseteq> \<langle>a \<inter> b\<rangle> q"
  using assms by(auto simp add: wp_ext)

lemma wp_ir_act_Int:
  assumes "a \<in> A"
      and "p \<subseteq> \<langle>a\<rangle> q"
    shows "p \<subseteq> \<langle>\<Inter> A\<rangle> q"
  using assms by(auto simp add: wp_ext)

lemma wp_ir_act_un:
  assumes "p \<subseteq> \<langle>a\<rangle> q"
      and "p \<subseteq> \<langle>b\<rangle> q"
    shows "p \<subseteq> \<langle>a \<union> b\<rangle> q"
  using assms by(auto simp add: wp_ext)

lemma wp_ir_act_un2:
  shows "p \<subseteq> \<langle>a \<union> b\<rangle> q \<longleftrightarrow> p \<subseteq> \<langle>b\<rangle> q \<and> p \<subseteq> \<langle>a\<rangle> q"
  by(auto simp add: wp_ext)

lemma wp_ir_act_Union:
  assumes "\<And> a. a \<in> A \<Longrightarrow> p \<subseteq> \<langle>a\<rangle> q"
    shows "p \<subseteq> \<langle>\<Union> A\<rangle> q"
  using assms by(auto simp add: wp_ext)

lemmas wp_ir_act = wp_ir_act_bot wp_ir_act_pred_mono wp_ir_act_act_mono wp_ir_act_int1
wp_ir_act_int2 wp_ir_act_Int wp_ir_act_un wp_ir_act_un2 wp_ir_act_Union

lemmas wp_pt = wp_ir_act wp_ir_pr

definition "wk_pres P = \<Union> {A . P \<subseteq> \<langle>A\<rangle> P}"

lemma wk_pres_spec:
  "P \<subseteq> \<langle>wk_pres P\<rangle> P"
  apply(simp add: wk_pres_def)
  by (metis (no_types, lifting) mem_Collect_eq wp_Sup2)

lemma wk_pres_intro:
  assumes "P \<subseteq> \<langle>A\<rangle> P"
  shows "A \<subseteq> wk_pres P"
  using assms by(fastforce simp add: wk_pres_def)

lemma int1_wk_pres:
  "P \<subseteq> \<langle>wk_pres P \<inter> wk_pres Q\<rangle> P"
  apply(rule wp_ir_act_int1)
  by (rule wk_pres_spec)

lemma int2_wk_pres:
  "Q \<subseteq> \<langle>wk_pres P \<inter> wk_pres Q\<rangle> Q"
  apply(rule wp_ir_act_int2)
  by (rule wk_pres_spec)

lemma Int_wk_pres:
  "P x \<subseteq> \<langle>\<Inter> x. wk_pres (P x)\<rangle> (P x)"
  apply(rule wp_ir_act_Int[where a="wk_pres (P x)"], simp)
  by (rule wk_pres_spec)

lemma Int_wk_pres_ext:
  "(\<Inter> x. wk_pres (P x)) = \<Union> {A . \<forall> x. P x \<subseteq> \<langle>A\<rangle> (P x)}"
  apply rule
  using Int_wk_pres[where P=P] apply fastforce
  by(fastforce simp add: wk_pres_def)


lemma emp_wp_univ [simp]:
   "\<langle>{}\<rangle> p = UNIV"
  by(auto simp add: wp_ext)

lemma wp_univ_univ [simp]:
  "\<langle>A\<rangle> UNIV = UNIV"
  by(auto simp add: wp_ext)

lemma wp_emp_emp [simp]:
  "-(Domain A) \<subseteq> \<langle>A\<rangle> {}"
  by(auto simp add: wp_ext)


lemma int_p_wp_diag:
  assumes "q \<subseteq> wp (p\<^sup>d) r"
  shows "q \<inter> p \<subseteq> r"
  using assms by(fastforce simp add: wp_ext diag_def)

lemma converse_int_p_wp_diag:
  assumes "q \<inter> p \<subseteq> r"
  shows "q \<subseteq> wp (p\<^sup>d) r"
  using assms by(fastforce simp add: wp_ext diag_def)

lemma converse_int_p_wp_diag_act:
  assumes "q \<inter> p \<subseteq> wp a r"
  shows "q \<subseteq> wp (p\<^sup>d O a) r"
  using assms by(fastforce simp add: wp_ext diag_def)

lemma wp_post_diag:
  "wp (a O p\<^sup>d) p = UNIV"
  by(fastforce simp add: wp_ext diag_def)

lemma wp_pointwise_discharge:
  "p \<subseteq> \<langle>a\<rangle> q  \<Longrightarrow> x \<in> p \<Longrightarrow> (x, y) \<in> a  \<Longrightarrow> y \<in> q"
  by(auto simp add: wp_ext)

lemma basic_wp_pres:
  assumes "Q \<subseteq> \<langle>A\<rangle> Q"
      and "P \<subseteq> Q"
    shows "P \<subseteq> \<langle>A\<rangle> Q"
  using assms(1) assms(2) by auto

lemma basic_wp_pres_disj:
  assumes "Q' \<subseteq> \<langle>A\<rangle> Q"
      and "P' \<subseteq> \<langle>A\<rangle> P"
    shows "P' \<union> Q' \<subseteq> \<langle>A\<rangle> (P \<union> Q)"
  by (metis Un_subset_iff assms(1) assms(2) sup_commute wp_ir_pr_disj_intro1)

end
